from PyQt5.QtCore import QDate, QDateTime, QLocale, QRegularExpression, Qt, QTime
from PyQt5.QtGui import QRegularExpressionValidator, QStandardItem, QStandardItemModel
from PyQt5.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDateTimeEdit,
    QDialog,
    QDialogButtonBox,
    QDoubleSpinBox,
    QFormLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QListView,
    QPlainTextEdit,
    QScrollArea,
    QSizePolicy,
    QSpacerItem,
    QSpinBox,
    QVBoxLayout,
    QWidget,
)

from trufont.controls.nameTabWidget import NameTabWidget
from trufont.objects import settings

# TODO: add placeholder text (fallbacks) and update on font.info changed
# TODO: add gasp/name entries


def TransparentScrollArea(parent=None):
    scrollArea = QScrollArea(parent)
    scrollArea.setFrameShape(scrollArea.NoFrame)
    scrollArea.setStyleSheet(".QScrollArea { background: transparent; }")
    scrollArea.viewport().setStyleSheet(".QWidget { background: transparent; }")
    return scrollArea


class FontInfoWindow(QDialog):
    def __init__(self, font, parent=None):
        super().__init__(parent)
        self.font = font

        self.tabWidget = NameTabWidget(self)
        self.tabWidget.addNamedTab(GeneralTab(self.font))
        self.tabWidget.addNamedTab(LegalTab(self.font))
        self.tabWidget.addNamedTab(OpenTypeTab(self.font))
        self.tabWidget.addNamedTab(PostScriptTab(self.font))

        buttonBox = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel, self
        )

        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)

        mainLayout = QVBoxLayout(self)
        mainLayout.addWidget(self.tabWidget)
        mainLayout.addWidget(buttonBox)

        self.setWindowTitle(
            self.tr("Font Info – {0} {1}").format(
                self.font.info.familyName, self.font.info.styleName
            )
        )

        self.readSettings()

    def readSettings(self):
        geometry = settings.fontInfoWindowGeometry()
        if geometry:
            self.restoreGeometry(geometry)

    def writeSettings(self):
        settings.setFontInfoWindowGeometry(self.saveGeometry())

    def accept(self):
        for i in range(self.tabWidget.count()):
            self.tabWidget.widget(i).storeValues()
        super().accept()

    def moveEvent(self, event):
        self.writeSettings()

    resizeEvent = moveEvent


class QSpinBoxZeroPadded(QSpinBox):
    def __init__(self, parent, digits):
        self.padding_digits = digits
        super().__init__(parent)

    def textFromValue(self, value):
        return str(value).zfill(self.padding_digits)


class TabWidget(QWidget):
    def __init__(self, font, parent=None, name=None):
        self.name = name
        self.font = font
        super().__init__(parent)

        self.loadFunc = {
            "s": self.loadString,
            "s+": self.loadMultilineString,
            "i": self.loadInteger,
            "pi": self.loadPositiveInteger,
            "zpi": self.loadZeroPaddedPositiveInteger,
            "if": self.loadIntegerFloat,
            "pif": self.loadPositiveIntegerFloat,
            "if+": self.loadIntegerFloatList,
            "b": self.loadBoolean,
        }
        self.storeFunc = {
            "s": self.storeString,
            "s+": self.storeMultilineString,
            "i": self.storeInteger,
            "pi": self.storePositiveInteger,
            "zpi": self.storePositiveInteger,
            "if": self.storeIntegerFloat,
            "pif": self.storePositiveIntegerFloat,
            "if+": self.storeIntegerFloatList,
            "b": self.storeBoolean,
        }
        self.attributeType = {}

    # Convenience method for loading simple optional attributes into a line
    # edit, spin box or boolean box.
    def loadCustomIntoForm(self, attribute, attrtype, labeltext, form):
        label = RCheckBox(labeltext)
        edit = self.load(attribute, attrtype)
        label.clicked.connect(edit.setEnabled)
        if getattr(self.font.info, attribute) is not None:
            label.setChecked(True)
        else:
            edit.setEnabled(False)
        form.addRow(label, edit)

        return edit

    def load(self, attribute, attrtype):
        if attrtype not in self.loadFunc:
            raise ValueError("`attrtype` must be listed in `self.loadFunc`.")

        self.attributeType[attribute] = attrtype
        edit = self.loadFunc[attrtype](attribute)

        return edit

    def store(self, attribute):
        if attribute not in self.attributeType:
            raise ValueError(
                "You can only store attributes that were " "previously `self.load()`ed."
            )

        attrtype = self.attributeType[attribute]
        self.storeFunc[attrtype](attribute)

    def loadString(self, attribute):
        value = getattr(self.font.info, attribute)
        edit = QLineEdit(value, self)
        setattr(self, attribute + "Edit", edit)
        return edit

    def storeString(self, attribute):
        edit = getattr(self, attribute + "Edit")
        value = None
        if edit.isEnabled():
            value = edit.text()
        setattr(self.font.info, attribute, value)

    def loadMultilineString(self, attribute):
        value = getattr(self.font.info, attribute)
        edit = QPlainTextEdit(value, self)
        setattr(self, attribute + "Edit", edit)
        return edit

    def storeMultilineString(self, attribute):
        edit = getattr(self, attribute + "Edit")
        value = None
        if edit.isEnabled():
            value = edit.toPlainText()
        setattr(self.font.info, attribute, value)

    def loadInteger(self, attribute):
        value = getattr(self.font.info, attribute)
        if value is None:
            value = 0
        edit = QSpinBox(self)
        edit.setRange(-2147483648, 2147483647)
        edit.setValue(value)
        setattr(self, attribute + "Edit", edit)
        return edit

    def storeInteger(self, attribute):
        edit = getattr(self, attribute + "Edit")
        value = edit.value()
        if edit.isEnabled():
            setattr(self.font.info, attribute, value)
        else:
            setattr(self.font.info, attribute, None)

    def loadPositiveInteger(self, attribute):
        value = getattr(self.font.info, attribute)
        if value is None:
            value = 0
        edit = QSpinBox(self)
        edit.setRange(0, 2147483647)
        edit.setValue(value)
        setattr(self, attribute + "Edit", edit)
        return edit

    def loadZeroPaddedPositiveInteger(self, attribute, digits=3):
        value = getattr(self.font.info, attribute)
        if value is None:
            value = 0
        edit = QSpinBoxZeroPadded(self, digits)
        edit.setRange(0, 2147483647)
        edit.setValue(value)
        setattr(self, attribute + "Edit", edit)
        return edit

    storePositiveInteger = storeInteger

    def loadIntegerFloat(self, attribute):
        value = getattr(self.font.info, attribute)
        if value is None:
            value = 0
        edit = FloatSpinBox(self)
        edit.setRange(-2147483648, 2147483647)
        edit.setDecimals(8)
        edit.setValue(value)
        setattr(self, attribute + "Edit", edit)
        return edit

    def storeIntegerFloat(self, attribute):
        edit = getattr(self, attribute + "Edit")
        value = edit.value()
        if edit.isEnabled():
            if value - int(value) == 0:  # Store ints as actual ints
                setattr(self.font.info, attribute, int(value))
            else:
                setattr(self.font.info, attribute, float(value))
        else:
            setattr(self.font.info, attribute, None)

    def loadPositiveIntegerFloat(self, attribute):
        value = getattr(self.font.info, attribute)
        if value is None:
            value = 0
        edit = FloatSpinBox(self)
        edit.setRange(0, 2147483647)
        edit.setDecimals(8)
        edit.setValue(value)
        setattr(self, attribute + "Edit", edit)
        return edit

    storePositiveIntegerFloat = storeIntegerFloat

    def loadIntegerFloatList(self, attribute):
        values = " ".join(str(val) for val in getattr(self.font.info, attribute))
        edit = QLineEdit(values, self)
        validator = QRegularExpressionValidator(self)
        validator.setRegularExpression(QRegularExpression(r"(-?\d+([,.]\d+)?\s*)*"))
        edit.setValidator(validator)
        setattr(self, attribute + "Edit", edit)
        return edit

    def storeIntegerFloatList(self, attribute):
        edit = getattr(self, attribute + "Edit")
        attributeValues = None
        if edit.isEnabled():
            values = edit.text().split()
            attributeValues = []
            for val in values:
                if "." in val:
                    attributeValues.append(float(val))
                elif "," in val:
                    v = val.replace(",", ".")
                    attributeValues.append(float(v))
                elif val not in ("", " "):
                    attributeValues.append(int(val))
        setattr(self.font.info, attribute, attributeValues)

    def loadBoolean(self, attribute):
        value = getattr(self.font.info, attribute)
        edit = QCheckBox(self)
        if value is None:
            edit.setChecked(False)
        else:
            edit.setChecked(value)
        setattr(self, attribute + "Edit", edit)
        return edit

    def storeBoolean(self, attribute):
        edit = getattr(self, attribute + "Edit")
        value = edit.isChecked()
        if edit.isEnabled():
            setattr(self.font.info, attribute, bool(value))
        else:
            setattr(self.font.info, attribute, None)


class GeneralTab(TabWidget):
    def __init__(self, font, parent=None):
        super().__init__(font, parent)
        self.name = self.tr("General")

        mainLayout = QVBoxLayout(self)

        # General metadata
        generalLayout = QHBoxLayout()

        # General general metadata
        g1FormLayout = QFormLayout()
        g1FormLayout.setLabelAlignment(
            Qt.AlignRight | Qt.AlignTrailing | Qt.AlignVCenter
        )

        familyNameLabel = QLabel(self.tr("Family name:"), self)
        edit = self.load("familyName", "s")
        g1FormLayout.addRow(familyNameLabel, edit)

        styleNameLabel = QLabel(self.tr("Style name:"), self)
        edit = self.load("styleName", "s")
        g1FormLayout.addRow(styleNameLabel, edit)

        styleMapLabel = RCheckBox(self.tr("Style map:"), self)
        styleMapLayout = QHBoxLayout()
        self.styleMapFamilyEdit = QLineEdit(font.info.styleMapFamilyName, self)
        self.styleMapStyleDrop = QComboBox(self)
        items = [
            self.tr("Regular"),
            self.tr("Italic"),
            self.tr("Bold"),
            self.tr("Bold Italic"),
        ]
        self.styleMapStyleDrop.insertItems(0, items)
        styleMapLabel.clicked.connect(self.styleMapFamilyEdit.setEnabled)
        styleMapLabel.clicked.connect(self.styleMapStyleDrop.setEnabled)
        sn = font.info.styleMapStyleName
        styleNames = {"regular": 0, "italic": 1, "bold": 2, "bold italic": 3}
        if sn and sn in styleNames:  # XXX: need sanity checks?
            styleMapLabel.setChecked(True)
            self.styleMapStyleDrop.setCurrentIndex(styleNames[sn])
        else:
            self.styleMapFamilyEdit.setEnabled(False)
            self.styleMapStyleDrop.setEnabled(False)
        styleMapLayout.addWidget(self.styleMapFamilyEdit)
        styleMapLayout.addWidget(self.styleMapStyleDrop)
        g1FormLayout.addRow(styleMapLabel, styleMapLayout)

        versionLabel = QLabel(self.tr("Version:"), self)
        versionLayout = QHBoxLayout()
        editMajor = self.load("versionMajor", "i")
        versionLayout.addWidget(editMajor)
        versionLayout.setStretch(0, 1)
        versionDotLabel = QLabel(".", self)
        versionLayout.addWidget(versionDotLabel)
        editMinor = self.load("versionMinor", "zpi")
        versionLayout.addWidget(editMinor)
        versionLayout.setStretch(2, 1)
        g1FormLayout.addRow(versionLabel, versionLayout)

        dateCreatedLabel = QLabel(self.tr("Date created:"), self)
        dateTime = QDateTime()
        # dateTime.fromString(font.info.openTypeHeadCreated, "yyyy/MM/dd
        # hh:mm:ss") # XXX: why does this not work?
        dateCreated = font.info.openTypeHeadCreated
        if dateCreated:
            parse = dateCreated.split(" ")
            if len(parse) == 2:
                date = parse[0].split("/")
                date = QDate(*(int(val) for val in date))
                dateTime.setDate(date)
                time = parse[1].split(":")
                time = QTime(*(int(val) for val in time))
                dateTime.setTime(time)
        else:
            cur = QDateTime.currentDateTime()
            dateTime.setDate(cur.date())
            dateTime.setTime(cur.time())
        self.dateCreatedEdit = QDateTimeEdit(dateTime, self)
        g1FormLayout.addRow(dateCreatedLabel, self.dateCreatedEdit)

        generalLayout.addLayout(g1FormLayout)
        generalLayout.setStretch(0, 1)

        # General metrics metadata
        g2FormLayout = QFormLayout()
        g2FormLayout.setLabelAlignment(
            Qt.AlignRight | Qt.AlignTrailing | Qt.AlignVCenter
        )

        unitsPerEmLabel = QLabel(self.tr("Units per em:"), self)
        edit = self.load("unitsPerEm", "pif")
        g2FormLayout.addRow(unitsPerEmLabel, edit)

        ascenderLabel = QLabel(self.tr("Ascender:"), self)
        edit = self.load("ascender", "if")
        g2FormLayout.addRow(ascenderLabel, edit)

        descenderLabel = QLabel(self.tr("Descender:"), self)
        edit = self.load("descender", "if")
        g2FormLayout.addRow(descenderLabel, edit)

        capHeightLabel = QLabel(self.tr("Cap height:"), self)
        edit = self.load("capHeight", "if")
        g2FormLayout.addRow(capHeightLabel, edit)

        xHeightLabel = QLabel(self.tr("x-height:"), self)
        edit = self.load("xHeight", "if")
        g2FormLayout.addRow(xHeightLabel, edit)

        italicAngleLabel = QLabel(self.tr("Italic angle:"), self)
        edit = self.load("italicAngle", "if")
        g2FormLayout.addRow(italicAngleLabel, edit)

        generalLayout.addLayout(g2FormLayout)

        mainLayout.addLayout(generalLayout)

        # Notes
        notesLayout = QFormLayout()
        notesLayout.setLabelAlignment(
            Qt.AlignRight | Qt.AlignTrailing | Qt.AlignVCenter
        )
        notesLayout.setRowWrapPolicy(QFormLayout.WrapAllRows)

        noteLabel = QLabel(self.tr("Notes:"), self)
        edit = self.load("note", "s+")
        noteSizePolicy = edit.sizePolicy()
        noteSizePolicy.setVerticalStretch(1)
        edit.setSizePolicy(noteSizePolicy)
        notesLayout.addRow(noteLabel, edit)

        mainLayout.addLayout(notesLayout)

    def storeValues(self):
        self.store("familyName")
        self.store("styleName")

        font = self.font
        if self.styleMapFamilyEdit.isEnabled() and self.styleMapStyleDrop.isEnabled():
            font.info.styleMapFamilyName = self.styleMapFamilyEdit.text()
            styleNames = ["regular", "italic", "bold", "bold italic"]
            sn = styleNames[self.styleMapStyleDrop.currentIndex()]
            font.info.styleMapStyleName = sn
        else:
            font.info.styleMapFamilyName = None
            font.info.styleMapStyleName = None

        self.store("versionMajor")
        self.store("versionMinor")

        font.info.openTypeHeadCreated = QLocale.c().toString(
            self.dateCreatedEdit.dateTime(), "yyyy/MM/dd hh:mm:ss"
        )

        self.store("unitsPerEm")
        self.store("italicAngle")
        self.store("ascender")
        self.store("descender")
        self.store("capHeight")
        self.store("xHeight")

        self.store("note")


class LegalTab(TabWidget):
    def __init__(self, font, parent=None):
        super().__init__(font, parent)
        self.name = self.tr("Legal")

        mainLayout = QFormLayout(self)
        mainLayout.setLabelAlignment(Qt.AlignRight | Qt.AlignTrailing | Qt.AlignVCenter)

        designerLabel = QLabel(self.tr("Designer:"), self)
        edit = self.load("openTypeNameDesigner", "s")
        mainLayout.addRow(designerLabel, edit)

        designerURLLabel = QLabel(self.tr("Designer URL:"), self)
        edit = self.load("openTypeNameDesignerURL", "s")
        mainLayout.addRow(designerURLLabel, edit)

        manufacturerLabel = QLabel(self.tr("Manufacturer:"), self)
        edit = self.load("openTypeNameManufacturer", "s")
        mainLayout.addRow(manufacturerLabel, edit)

        manufacturerURLLabel = QLabel(self.tr("Manufacturer URL:"), self)
        edit = self.load("openTypeNameManufacturerURL", "s")
        mainLayout.addRow(manufacturerURLLabel, edit)

        copyrightLabel = QLabel(self.tr("Copyright:"), self)
        edit = self.load("copyright", "s+")
        copyrightSizePolicy = edit.sizePolicy()
        copyrightSizePolicy.setVerticalStretch(1)
        edit.setSizePolicy(copyrightSizePolicy)
        mainLayout.addRow(copyrightLabel, edit)

        licenseLabel = QLabel(self.tr("License:"), self)
        edit = self.load("openTypeNameLicense", "s+")
        licenseSizePolicy = edit.sizePolicy()
        licenseSizePolicy.setVerticalStretch(1)
        edit.setSizePolicy(licenseSizePolicy)
        mainLayout.addRow(licenseLabel, edit)

        licenseURLLabel = QLabel(self.tr("License URL:"), self)
        edit = self.load("openTypeNameLicenseURL", "s")
        mainLayout.addRow(licenseURLLabel, edit)

        trademarkLabel = QLabel(self.tr("Trademark:"), self)
        edit = self.load("trademark", "s")
        mainLayout.addRow(trademarkLabel, edit)

    def storeValues(self):
        self.store("openTypeNameDesigner")
        self.store("openTypeNameDesignerURL")
        self.store("openTypeNameManufacturer")
        self.store("openTypeNameManufacturerURL")
        self.store("copyright")
        self.store("openTypeNameLicense")
        self.store("openTypeNameLicenseURL")
        self.store("trademark")


class OpenTypeTab(TabWidget):
    def __init__(self, font, parent=None):
        super().__init__(font, parent)
        self.name = self.tr("OpenType")

        mainLayout = QVBoxLayout(self)
        mainLayout.setContentsMargins(0, 0, 0, 0)

        tableScrollArea = TransparentScrollArea(self)
        tableScrollArea.setWidgetResizable(True)
        tableArea = QWidget(tableScrollArea)
        tableScrollArea.setWidget(tableArea)
        mainLayout.addWidget(tableScrollArea)

        tableLayout = QVBoxLayout(tableArea)

        tableLayout.addWidget(self.setupHeadGroup())
        tableLayout.addWidget(self.setupNameGroup())
        tableLayout.addWidget(self.setupHheaGroup())
        tableLayout.addWidget(self.setupVheaGroup())
        tableLayout.addWidget(self.setupOs2Group())

        spacer = QSpacerItem(10, 20, QSizePolicy.Minimum, QSizePolicy.Expanding)
        tableLayout.addItem(spacer)

    def setupHeadGroup(self):
        headGroup = QGroupBox(self.tr("head table"))
        headLayout = QVBoxLayout(headGroup)

        headAttributesLayout = QFormLayout()
        headAttributesLayout.setLabelAlignment(
            Qt.AlignRight | Qt.AlignTrailing | Qt.AlignVCenter
        )

        self.loadCustomIntoForm(
            "openTypeHeadLowestRecPPEM",
            "pi",
            self.tr("Lowest rec. PPEM:"),
            headAttributesLayout,
        )

        flagsLabel = RCheckBox(self.tr("Flags:"))
        self.flagsEdit = BitListView(self)
        flagsLabel.clicked.connect(self.flagsEdit.setEnabled)
        flagItems = [
            self.tr("Bit 0: Baseline for font at y=0."),
            self.tr("Bit 1: Left sidebearing point at x=0."),
            self.tr("Bit 2: Instructions may depend on point size."),
            self.tr(
                "Bit 3: Force ppem to integer values for all " "internal scaler math."
            ),
            self.tr("Bit 4: Instructions may alter advance width."),
            self.tr("Bit 11: Font data is 'lossless'."),
            self.tr("Bit 12: Font converted."),
            self.tr("Bit 13: Font optimized for ClearType™."),
            self.tr("Bit 14: Last Resort font."),
        ]
        itemToFlagMap = {0: 0, 1: 1, 2: 2, 3: 3, 4: 4, 5: 11, 6: 12, 7: 13, 8: 14}
        flags = self.font.info.openTypeHeadFlags
        model = QStandardItemModel(9, 1)
        for index, elem in enumerate(flagItems):
            item = QStandardItem()
            item.setText(elem)
            item.setCheckable(True)
            bit = itemToFlagMap[index]
            if flags is not None and bit in flags:
                item.setCheckState(Qt.Checked)
            else:
                item.setCheckState(Qt.Unchecked)
            model.setItem(index, item)
        self.flagsEdit.setModel(model)
        if flags is not None:
            flagsLabel.setChecked(True)
        else:
            self.flagsEdit.setEnabled(False)
        headAttributesLayout.addRow(flagsLabel, self.flagsEdit)

        headLayout.addLayout(headAttributesLayout)

        return headGroup

    def setupNameGroup(self):
        nameGroup = QGroupBox(self.tr("name table"))
        nameLayout = QVBoxLayout(nameGroup)

        nameAttributesLayout = QFormLayout()
        nameAttributesLayout.setLabelAlignment(
            Qt.AlignRight | Qt.AlignTrailing | Qt.AlignVCenter
        )

        self.loadCustomIntoForm(
            "openTypeNamePreferredFamilyName",
            "s",
            self.tr("Pref. family name:"),
            nameAttributesLayout,
        )
        self.loadCustomIntoForm(
            "openTypeNamePreferredSubfamilyName",
            "s",
            self.tr("Pref. subfamily name:"),
            nameAttributesLayout,
        )
        self.loadCustomIntoForm(
            "openTypeNameCompatibleFullName",
            "s",
            self.tr("Compatible full name:"),
            nameAttributesLayout,
        )
        self.loadCustomIntoForm(
            "openTypeNameWWSFamilyName",
            "s",
            self.tr("WWS family name:"),
            nameAttributesLayout,
        )
        self.loadCustomIntoForm(
            "openTypeNameWWSSubfamilyName",
            "s",
            self.tr("WWS subfamily name:"),
            nameAttributesLayout,
        )
        self.loadCustomIntoForm(
            "openTypeNameVersion", "s", self.tr("Version:"), nameAttributesLayout
        )
        self.loadCustomIntoForm(
            "openTypeNameUniqueID", "s", self.tr("Unique ID:"), nameAttributesLayout
        )
        self.loadCustomIntoForm(
            "openTypeNameDescription",
            "s",
            self.tr("Description:"),
            nameAttributesLayout,
        )
        self.loadCustomIntoForm(
            "openTypeNameSampleText", "s", self.tr("Sample text:"), nameAttributesLayout
        )

        nameLayout.addLayout(nameAttributesLayout)

        return nameGroup

    def setupHheaGroup(self):
        hheaGroup = QGroupBox(self.tr("hhea table"))
        hheaLayout = QVBoxLayout(hheaGroup)

        hheaAttributesLayout = QFormLayout()
        hheaAttributesLayout.setLabelAlignment(
            Qt.AlignRight | Qt.AlignTrailing | Qt.AlignVCenter
        )

        self.loadCustomIntoForm(
            "openTypeHheaAscender", "i", self.tr("Ascender:"), hheaAttributesLayout
        )
        self.loadCustomIntoForm(
            "openTypeHheaDescender", "i", self.tr("Descender:"), hheaAttributesLayout
        )
        self.loadCustomIntoForm(
            "openTypeHheaLineGap", "i", self.tr("LineGap:"), hheaAttributesLayout
        )
        self.loadCustomIntoForm(
            "openTypeHheaCaretSlopeRise",
            "i",
            self.tr("caretSlopeRise:"),
            hheaAttributesLayout,
        )
        self.loadCustomIntoForm(
            "openTypeHheaCaretSlopeRun",
            "i",
            self.tr("caretSlopeRun:"),
            hheaAttributesLayout,
        )
        self.loadCustomIntoForm(
            "openTypeHheaCaretOffset",
            "i",
            self.tr("caretOffset:"),
            hheaAttributesLayout,
        )

        hheaLayout.addLayout(hheaAttributesLayout)

        return hheaGroup

    def setupVheaGroup(self):
        vheaGroup = QGroupBox(self.tr("vhea table"))
        vheaLayout = QVBoxLayout(vheaGroup)

        vheaAttributesLayout = QFormLayout()
        vheaAttributesLayout.setLabelAlignment(
            Qt.AlignRight | Qt.AlignTrailing | Qt.AlignVCenter
        )

        self.loadCustomIntoForm(
            "openTypeVheaVertTypoAscender",
            "i",
            self.tr("V. Ascender:"),
            vheaAttributesLayout,
        )
        self.loadCustomIntoForm(
            "openTypeVheaVertTypoDescender",
            "i",
            self.tr("V. Descender:"),
            vheaAttributesLayout,
        )
        self.loadCustomIntoForm(
            "openTypeVheaVertTypoLineGap",
            "i",
            self.tr("V. LineGap:"),
            vheaAttributesLayout,
        )
        self.loadCustomIntoForm(
            "openTypeVheaCaretSlopeRise",
            "i",
            self.tr("V. caretSlopeRise:"),
            vheaAttributesLayout,
        )
        self.loadCustomIntoForm(
            "openTypeVheaCaretSlopeRun",
            "i",
            self.tr("V. caretSlopeRun:"),
            vheaAttributesLayout,
        )
        self.loadCustomIntoForm(
            "openTypeVheaCaretOffset",
            "i",
            self.tr("V. caretOffset:"),
            vheaAttributesLayout,
        )

        vheaLayout.addLayout(vheaAttributesLayout)

        return vheaGroup

    def setupOs2Group(self):
        font = self.font
        os2Group = QGroupBox(self.tr("os2 table"))
        os2Layout = QVBoxLayout(os2Group)

        os2AttributesLayout = QFormLayout()
        os2AttributesLayout.setLabelAlignment(
            Qt.AlignRight | Qt.AlignTrailing | Qt.AlignVCenter
        )

        self.loadCustomIntoForm(
            "openTypeOS2VendorID", "s", self.tr("Vendor ID:"), os2AttributesLayout
        ).setMaxLength(4)

        usWidthClassLabel = RCheckBox(self.tr("Width class:"))
        self.usWidthClassDrop = QComboBox(self)
        items = [
            self.tr("Ultra-condensed"),
            self.tr("Extra-condensed"),
            self.tr("Condensed"),
            self.tr("Semi-Condensed"),
            self.tr("Medium (normal)"),
            self.tr("Semi-expanded"),
            self.tr("Expanded"),
            self.tr("Extra-expanded"),
            self.tr("Ultra-expanded"),
        ]
        self.usWidthClassDrop.insertItems(0, items)
        usWidthClassLabel.clicked.connect(self.usWidthClassDrop.setEnabled)
        if font.info.openTypeOS2WidthClass:
            usWidthClassLabel.setChecked(True)
            self.usWidthClassDrop.setCurrentIndex(font.info.openTypeOS2WidthClass - 1)
        else:
            self.usWidthClassDrop.setEnabled(False)
            self.usWidthClassDrop.setCurrentIndex(4)
        os2AttributesLayout.addRow(usWidthClassLabel, self.usWidthClassDrop)

        usWeightClassLabel = RCheckBox(self.tr("Weight class:"))
        self.usWeightClassDrop = QComboBox(self)
        items = [
            self.tr("Thin"),
            self.tr("Extra-light (Ultra-light)"),
            self.tr("Light"),
            self.tr("Normal (Regular)"),
            self.tr("Medium"),
            self.tr("Semi-bold (Demi-bold)"),
            self.tr("Bold"),
            self.tr("Extra-bold (Ultra-bold)"),
            self.tr("Black (Heavy)"),
        ]
        self.usWeightClassDrop.insertItems(0, items)
        usWeightClassLabel.clicked.connect(self.usWeightClassDrop.setEnabled)
        wc = font.info.openTypeOS2WeightClass
        if wc and wc % 100 == 0:
            usWeightClassLabel.setChecked(True)
            self.usWeightClassDrop.setCurrentIndex(
                font.info.openTypeOS2WeightClass / 100 - 1
            )
        else:
            self.usWeightClassDrop.setEnabled(False)
            self.usWeightClassDrop.setCurrentIndex(3)
        os2AttributesLayout.addRow(usWeightClassLabel, self.usWeightClassDrop)

        fsSelectionLabel = RCheckBox(self.tr("fsSelection:"))
        fsSelection = font.info.openTypeOS2Selection
        self.fsSelectionList = BitListView(self)
        self.fsSelectionList.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        fsSelectionLabel.clicked.connect(self.fsSelectionList.setEnabled)
        items = [
            self.tr("Bit 1 UNDERSCORE"),
            self.tr("Bit 2 NEGATIVE"),
            self.tr("Bit 3 OUTLINED"),
            self.tr("Bit 4 STRIKEOUT"),
            self.tr("Bit 7 USE_TYPO_METRICS"),
            self.tr("Bit 8 WWS"),
            self.tr("Bit 9 OBLIQUE"),
        ]
        itemToFlagMap = {0: 1, 1: 2, 2: 3, 3: 4, 4: 7, 5: 8, 6: 9}
        # http://stackoverflow.com/a/26613163
        model = QStandardItemModel(7, 1)
        for index, elem in enumerate(items):
            item = QStandardItem()
            item.setText(elem)
            item.setCheckable(True)
            bit = itemToFlagMap[index]
            if fsSelection is not None and bit in fsSelection:
                item.setCheckState(Qt.Checked)
            else:
                item.setCheckState(Qt.Unchecked)
            model.setItem(index, item)
        self.fsSelectionList.setModel(model)
        if fsSelection is not None:
            fsSelectionLabel.setChecked(True)
        else:
            self.fsSelectionList.setEnabled(False)
        os2AttributesLayout.addRow(fsSelectionLabel, self.fsSelectionList)

        # XXX: disabled QCheckBoxes eat scroll events?! Can't scroll past
        # disabled checkboxes unless outside their rectangle.
        fsTypeLabel = RCheckBox(self.tr("Embedding:"))
        fsType = font.info.openTypeOS2Type
        self.fsTypeDrop = QComboBox(self)
        items = [
            self.tr("No embedding restrictions"),
            self.tr("Restricted embedding"),
            self.tr("Preview and print embedding allowed"),
            self.tr("Editable embedding allowed"),
        ]
        self.noSubsettingBox = QCheckBox(self.tr("No subsetting"))
        self.bitmapEmbeddingOnlyBox = QCheckBox(self.tr("Bitmap embedding only"))
        fsTypeLabel.clicked.connect(self.fsTypeDrop.setEnabled)
        fsTypeLabel.clicked.connect(self.noSubsettingBox.setEnabled)
        fsTypeLabel.clicked.connect(self.bitmapEmbeddingOnlyBox.setEnabled)
        fsTypeLabel.clicked.connect(
            lambda: self._updateFsTypeVisibility(self.fsTypeDrop.currentIndex())
        )
        self.fsTypeDrop.currentIndexChanged[int].connect(self._updateFsTypeVisibility)
        self.fsTypeDrop.insertItems(0, items)
        if fsType is not None:
            fsTypeLabel.setChecked(True)
            for i in range(1, 4):
                if i in fsType:
                    self.fsTypeDrop.setCurrentIndex(i)
                    break
            self.noSubsettingBox.setChecked(8 in fsType)
            self.bitmapEmbeddingOnlyBox.setChecked(9 in fsType)
        else:
            self.fsTypeDrop.setEnabled(False)
        embeddingLayout = QVBoxLayout()
        embeddingLayout.addWidget(self.fsTypeDrop)
        embeddingLayout.addWidget(self.noSubsettingBox)
        embeddingLayout.addWidget(self.bitmapEmbeddingOnlyBox)
        os2AttributesLayout.addRow(fsTypeLabel, embeddingLayout)

        panoseLabel = RCheckBox(self.tr("PANOSE:"))
        panose = font.info.openTypeOS2Panose
        panoseLayout = QFormLayout()
        panoseLayout.setLabelAlignment(
            Qt.AlignRight | Qt.AlignTrailing | Qt.AlignVCenter
        )
        panoseFamilyLabel = QLabel(self.tr("Family:"))
        panoseFamilyTypes = [
            self.tr("Any"),
            self.tr("No fit"),
            self.tr("Text and display"),
            self.tr("Script"),
            self.tr("Decorative"),
            self.tr("Pictorial"),
        ]
        self.panoseFamilyDrop = QComboBox(self)
        self.panoseFamilyDrop.insertItems(0, panoseFamilyTypes)
        panoseLayout.addRow(panoseFamilyLabel, self.panoseFamilyDrop)
        panoseSerifsLabel = QLabel(self.tr("Serifs:"))
        panoseSerifsTypes = [
            self.tr("Any"),
            self.tr("No fit"),
            self.tr("Cove"),
            self.tr("Obtuse cove"),
            self.tr("Square cove"),
            self.tr("Obtuse square cove"),
            self.tr("Square"),
            self.tr("Thin"),
            self.tr("Bone"),
            self.tr("Exaggerated"),
            self.tr("Triangle"),
            self.tr("Normal sans-serif"),
            self.tr("Obtuse sans-serif"),
            self.tr("Perp sans-serif"),
            self.tr("Flared"),
            self.tr("Rounded"),
        ]
        self.panoseSerifsDrop = QComboBox(self)
        self.panoseSerifsDrop.insertItems(0, panoseSerifsTypes)
        panoseLayout.addRow(panoseSerifsLabel, self.panoseSerifsDrop)
        panoseWeightLabel = QLabel(self.tr("Weight:"))
        panoseWeightTypes = [
            self.tr("Any"),
            self.tr("No fit"),
            self.tr("Very light"),
            self.tr("Light"),
            self.tr("Thin"),
            self.tr("Book"),
            self.tr("Medium"),
            self.tr("Demibold"),
            self.tr("Bold"),
            self.tr("Heavy"),
            self.tr("Black"),
            self.tr("Nord"),
        ]
        self.panoseWeightDrop = QComboBox(self)
        self.panoseWeightDrop.insertItems(0, panoseWeightTypes)
        panoseLayout.addRow(panoseWeightLabel, self.panoseWeightDrop)
        panoseProportionLabel = QLabel(self.tr("Proportion:"))
        panoseProportionTypes = [
            self.tr("Any"),
            self.tr("No fit"),
            self.tr("Old style"),
            self.tr("Modern"),
            self.tr("Even width"),
            self.tr("Expanded"),
            self.tr("Condensed"),
            self.tr("Very expanded"),
            self.tr("Very condensed"),
            self.tr("Monospaced"),
        ]
        self.panoseProportionDrop = QComboBox(self)
        self.panoseProportionDrop.insertItems(0, panoseProportionTypes)
        panoseLayout.addRow(panoseProportionLabel, self.panoseProportionDrop)
        panoseContrastLabel = QLabel(self.tr("Contrast:"))
        panoseContrastTypes = [
            self.tr("Any"),
            self.tr("No fit"),
            self.tr("None"),
            self.tr("Very low"),
            self.tr("Low"),
            self.tr("Medium low"),
            self.tr("Medium"),
            self.tr("Medium high"),
            self.tr("High"),
            self.tr("Very high"),
        ]
        self.panoseContrastDrop = QComboBox(self)
        self.panoseContrastDrop.insertItems(0, panoseContrastTypes)
        panoseLayout.addRow(panoseContrastLabel, self.panoseContrastDrop)
        panoseStrokeVariationLabel = QLabel(self.tr("Stroke variation:"))
        panoseStrokeVariationTypes = [
            self.tr("Any"),
            self.tr("No fit"),
            self.tr("Gradual/diagonal"),
            self.tr("Gradual/transitional"),
            self.tr("Gradual/vertical"),
            self.tr("Gradual/horizontal"),
            self.tr("Rapid/vertical"),
            self.tr("Rapid/horizontal"),
            self.tr("Instant/vertical"),
        ]
        self.panoseStrokeVariationDrop = QComboBox(self)
        self.panoseStrokeVariationDrop.insertItems(0, panoseStrokeVariationTypes)
        panoseLayout.addRow(panoseStrokeVariationLabel, self.panoseStrokeVariationDrop)
        panoseArmStyleLabel = QLabel(self.tr("Arm style:"))
        panoseArmStyleTypes = [
            self.tr("Any"),
            self.tr("No fit"),
            self.tr("Straight arms/horizontal"),
            self.tr("Straight arms/wedge"),
            self.tr("Straight arms/vertical"),
            self.tr("Straight arms/single-serif"),
            self.tr("Straight arms/double-serif"),
            self.tr("Nonstraight arms/horizontal"),
            self.tr("Nonstraight arms/wedge"),
            self.tr("Nonstraight arms/vertical"),
            self.tr("Nonstraight arms/single-serif"),
            self.tr("Nonstraight arms/double-serif"),
        ]
        self.panoseArmStyleDrop = QComboBox(self)
        self.panoseArmStyleDrop.insertItems(0, panoseArmStyleTypes)
        panoseLayout.addRow(panoseArmStyleLabel, self.panoseArmStyleDrop)
        panoseLetterformLabel = QLabel(self.tr("Letterform:"))
        panoseLetterformTypes = [
            self.tr("Any"),
            self.tr("No fit"),
            self.tr("Normal/contact"),
            self.tr("Normal/weighted"),
            self.tr("Normal/boxed"),
            self.tr("Normal/flattened"),
            self.tr("Normal/rounded"),
            self.tr("Normal/off center"),
            self.tr("Normal/square"),
            self.tr("Oblique/contact"),
            self.tr("Oblique/weighted"),
            self.tr("Oblique/boxed"),
            self.tr("Oblique/flattened"),
            self.tr("Oblique/rounded"),
            self.tr("Oblique/off center"),
            self.tr("Oblique/square"),
        ]
        self.panoseLetterformDrop = QComboBox(self)
        self.panoseLetterformDrop.insertItems(0, panoseLetterformTypes)
        panoseLayout.addRow(panoseLetterformLabel, self.panoseLetterformDrop)
        panoseMidlineLabel = QLabel(self.tr("Midline:"))
        panoseMidlineTypes = [
            self.tr("Any"),
            self.tr("No fit"),
            self.tr("Standard/trimmed"),
            self.tr("Standard/pointed"),
            self.tr("Standard/serifed"),
            self.tr("High/trimmed"),
            self.tr("High/pointed"),
            self.tr("High/serifed"),
            self.tr("Constant/trimmed"),
            self.tr("Constant/pointed"),
            self.tr("Constant/serifed"),
            self.tr("Low/trimmed"),
            self.tr("Low/pointed"),
            self.tr("Low/serifed"),
        ]
        self.panoseMidlineDrop = QComboBox(self)
        self.panoseMidlineDrop.insertItems(0, panoseMidlineTypes)
        panoseLayout.addRow(panoseMidlineLabel, self.panoseMidlineDrop)
        panoseXHeightLabel = QLabel(self.tr("x-Height:"))
        panoseXHeightTypes = [
            self.tr("Any"),
            self.tr("No fit"),
            self.tr("Constant/small"),
            self.tr("Constant/standard"),
            self.tr("Constant/large"),
            self.tr("Ducking/small"),
            self.tr("Ducking/standard"),
            self.tr("Ducking/large"),
        ]
        self.panoseXHeightDrop = QComboBox(self)
        self.panoseXHeightDrop.insertItems(0, panoseXHeightTypes)
        panoseLayout.addRow(panoseXHeightLabel, self.panoseXHeightDrop)
        panoseLabel.clicked.connect(self.panoseFamilyDrop.setEnabled)
        panoseLabel.clicked.connect(self.panoseSerifsDrop.setEnabled)
        panoseLabel.clicked.connect(self.panoseWeightDrop.setEnabled)
        panoseLabel.clicked.connect(self.panoseProportionDrop.setEnabled)
        panoseLabel.clicked.connect(self.panoseContrastDrop.setEnabled)
        panoseLabel.clicked.connect(self.panoseStrokeVariationDrop.setEnabled)
        panoseLabel.clicked.connect(self.panoseArmStyleDrop.setEnabled)
        panoseLabel.clicked.connect(self.panoseLetterformDrop.setEnabled)
        panoseLabel.clicked.connect(self.panoseMidlineDrop.setEnabled)
        panoseLabel.clicked.connect(self.panoseXHeightDrop.setEnabled)
        if panose:
            panoseLabel.setChecked(True)
            self.panoseFamilyDrop.setCurrentIndex(panose[0])
            self.panoseSerifsDrop.setCurrentIndex(panose[1])
            self.panoseWeightDrop.setCurrentIndex(panose[2])
            self.panoseProportionDrop.setCurrentIndex(panose[3])
            self.panoseContrastDrop.setCurrentIndex(panose[4])
            self.panoseStrokeVariationDrop.setCurrentIndex(panose[5])
            self.panoseArmStyleDrop.setCurrentIndex(panose[6])
            self.panoseLetterformDrop.setCurrentIndex(panose[7])
            self.panoseMidlineDrop.setCurrentIndex(panose[8])
            self.panoseXHeightDrop.setCurrentIndex(panose[9])
        else:
            self.panoseFamilyDrop.setEnabled(False)
            self.panoseSerifsDrop.setEnabled(False)
            self.panoseWeightDrop.setEnabled(False)
            self.panoseProportionDrop.setEnabled(False)
            self.panoseContrastDrop.setEnabled(False)
            self.panoseStrokeVariationDrop.setEnabled(False)
            self.panoseArmStyleDrop.setEnabled(False)
            self.panoseLetterformDrop.setEnabled(False)
            self.panoseMidlineDrop.setEnabled(False)
            self.panoseXHeightDrop.setEnabled(False)
        os2AttributesLayout.addRow(panoseLabel, panoseLayout)

        unicodeRangesLabel = RCheckBox(self.tr("Unicode ranges:"))
        unicodeRangesItems = [
            self.tr("Basic Latin (0000–007F)"),
            self.tr("Latin-1 Supplement (0080–00FF)"),
            self.tr("Latin Extended-A (0100–017F)"),
            self.tr("Latin Extended-B (0180–024F)"),
            self.tr("IPA Extensions (0250–02AF, 1D00–1D7F, 1D80–1DBF)"),
            self.tr(
                "Spacing Modifier Letters, Modifier Tone Letters "
                "(02B0–02FF, A700–A71F)"
            ),
            self.tr("Combining Diacritical Marks (0300–036F, 1DC0–1DFF)"),
            self.tr("Greek and Coptic (0370–03FF)"),
            self.tr("Coptic (2C80–2CFF)"),
            self.tr(
                "Cyrillic and extensions (0400–04FF, 0500–052F, "
                "2DE0–2DFF, A640–A69F)"
            ),
            self.tr("Armenian (0530–058F)"),
            self.tr("Hebrew (0590–05FF)"),
            self.tr("Vai (A500–A63F)"),
            self.tr("Arabic (0600–06FF, 0750–077F)"),
            self.tr("NKo (07C0–07FF)"),
            self.tr("Devanagari (0900–097F)"),
            self.tr("Bengali (0980–09FF)"),
            self.tr("Gurmukhi (0A00–0A7F)"),
            self.tr("Gujarati (0A80–0AFF)"),
            self.tr("Oriya (0B00–0B7F)"),
            self.tr("Tamil (0B80–0BFF)"),
            self.tr("Telugu (0C00–0C7F)"),
            self.tr("Kannada (0C80–0CFF)"),
            self.tr("Malayalam (0D00–0D7F)"),
            self.tr("Thai (0E00–0E7F)"),
            self.tr("Lao (0E80–0EFF)"),
            self.tr("Georgian (10A0–10FF, 2D00–2D2F)"),
            self.tr("Balinese (1B00–1B7F)"),
            self.tr("Hangul Jamo (1100–11FF)"),
            self.tr("Latin Extended Additional (1E00–1EFF, 2C60–2C7F, " "A720–A7FF)"),
            self.tr("Greek Extended (1F00–1FFF)"),
            self.tr("Punctuation (2000–206F, 2E00–2E7F)"),
            self.tr("Superscripts And Subscripts (2070–209F)"),
            self.tr("Currency Symbols (20A0–20CF)"),
            self.tr("Combining Diacritical Marks For Symbols " "(20D0–20FF)"),
            self.tr("Letterlike Symbols (2100–214F)"),
            self.tr("Number Forms (2150–218F)"),
            self.tr("Arrows (2190–21FF, 27F0–27FF, 2900–297F, " "2B00–2BFF)"),
            self.tr(
                "Mathematical Operators (2200–22FF, 2A00–2AFF, " "27C0–27EF, 2980–29FF)"
            ),
            self.tr("Miscellaneous Technical (2300–23FF)"),
            self.tr("Control Pictures (2400–243F)"),
            self.tr("Optical Character Recognition (2440–245F)"),
            self.tr("Enclosed Alphanumerics (2460–24FF)"),
            self.tr("Box Drawing (2500–257F)"),
            self.tr("Block Elements (2580–259F)"),
            self.tr("Geometric Shapes (25A0–25FF)"),
            self.tr("Miscellaneous Symbols (2600–26FF)"),
            self.tr("Dingbats (2700–27BF)"),
            self.tr("CJK Symbols And Punctuation (3000–303F)"),
            self.tr("Hiragana (3040–309F)"),
            self.tr("Katakana (30A0–30FF, 31F0–31FF)"),
            self.tr("Bopomofo (3100–312F, 31A0–31BF)"),
            self.tr("Hangul Compatibility Jamo (3130–318F)"),
            self.tr("Phags-pa (A840–A87F)"),
            self.tr("Enclosed CJK Letters And Months (3200–32FF)"),
            self.tr("CJK Compatibility (3300–33FF)"),
            self.tr("Hangul Syllables (AC00–D7AF)"),
            self.tr("Non-Plane 0 * (D800–DFFF)"),
            self.tr("Phoenician (10900–1091F)"),
            self.tr(
                "CJK Unified Ideographs (2E80–2EFF, 2F00–2FDF, "
                "2FF0–2FFF, 3190–319F, 3400–4DBF, 4E00–9FFF, "
                "20000–2A6DF)"
            ),
            self.tr("Private Use Area (plane 0) (E000–F8FF)"),
            self.tr("CJK Strokes (31C0–31EF, F900–FAFF, 2F800–2FA1F)"),
            self.tr("Alphabetic Presentation Forms (FB00–FB4F)"),
            self.tr("Arabic Presentation Forms-A (FB50–FDFF)"),
            self.tr("Combining Half Marks (FE20–FE2F)"),
            self.tr("CJK Vertical Forms (FE10–FE1F, FE30–FE4F)"),
            self.tr("Small Form Variants (FE50–FE6F)"),
            self.tr("Arabic Presentation Forms-B (FE70–FEFF)"),
            self.tr("Halfwidth And Fullwidth Forms (FF00–FFEF)"),
            self.tr("Specials (FFF0–FFFF)"),
            self.tr("Tibetan (0F00–0FFF)"),
            self.tr("Syriac (0700–074F)"),
            self.tr("Thaana (0780–07BF)"),
            self.tr("Sinhala (0D80–0DFF)"),
            self.tr("Myanmar (1000–109F)"),
            self.tr("Ethiopic (1200–137F, 1380–139F, 2D80–2DDF)"),
            self.tr("Cherokee (13A0–13FF)"),
            self.tr("Unified Canadian Aboriginal Syllabics (1400–167F)"),
            self.tr("Ogham (1680–169F)"),
            self.tr("Runic (16A0–16FF)"),
            self.tr("Khmer (1780–17FF, 19E0–19FF)"),
            self.tr("Mongolian (1800–18AF)"),
            self.tr("Braille Patterns (2800–28FF)"),
            self.tr("Yi Syllables (A000–A48F, A490–A4CF)"),
            self.tr(
                "Tagalog, Hanunoo, Buhid, Tagbanwa (1700–171F, "
                "1720–173F, 1740–175F, 1760–177F)"
            ),
            self.tr("Old Italic (10300–1032F)"),
            self.tr("Gothic (10330–1034F)"),
            self.tr("Deseret (10400–1044F)"),
            self.tr(
                "Greek and Byzantine Musical Symbols (1D000–1D0FF, "
                "1D100–1D1FF, 1D200–1D24F)"
            ),
            self.tr("Mathematical Alphanumeric Symbols (1D400–1D7FF)"),
            self.tr("Private Use (planes 15 and 16) " "(FF000–FFFFD, 100000–10FFFD)"),
            self.tr("Variation Selectors (FE00–FE0F, E0100–E01EF)"),
            self.tr("Tags (E0000–E007F)"),
            self.tr("Limbu (1900–194F)"),
            self.tr("Tai Le (1950–197F)"),
            self.tr("New Tai Lue (1980–19DF)"),
            self.tr("Buginese (1A00–1A1F)"),
            self.tr("Glagolitic (2C00–2C5F)"),
            self.tr("Tifinagh (2D30–2D7F)"),
            self.tr("Yijing Hexagram Symbols (4DC0–4DFF)"),
            self.tr("Syloti Nagri (A800–A82F)"),
            self.tr(
                "Linear B, Aegean Numbers (10000–1007F, " "10080–100FF, 10100–1013F)"
            ),
            self.tr("Ancient Greek Numbers (10140–1018F)"),
            self.tr("Ugaritic (10380–1039F)"),
            self.tr("Old Persian (103A0–103DF)"),
            self.tr("Shavian (10450–1047F)"),
            self.tr("Osmanya (10480–104AF)"),
            self.tr("Cypriot Syllabary (10800–1083F)"),
            self.tr("Kharoshthi (10A00–10A5F)"),
            self.tr("Tai Xuan Jing Symbols (1D300–1D35F)"),
            self.tr("Cuneiform (12000–123FF, 12400–1247F)"),
            self.tr("Counting Rod Numerals (1D360–1D37F)"),
            self.tr("Sundanese (1B80–1BBF)"),
            self.tr("Lepcha (1C00–1C4F)"),
            self.tr("Ol Chiki (1C50–1C7F)"),
            self.tr("Saurashtra (A880–A8DF)"),
            self.tr("Kayah Li (A900–A92F)"),
            self.tr("Rejang (A930–A95F)"),
            self.tr("Cham (AA00–AA5F)"),
            self.tr("Ancient Symbols (10190–101CF)"),
            self.tr("Phaistos Disc (101D0–101FF)"),
            self.tr(
                "Carian, Lycian, Lydian (102A0–102DF, 10280–1029F, " "10920–1093F)"
            ),
            self.tr("Domino tiles, Mahjong tiles (1F030–1F09F, " "1F000–1F02F)"),
        ]
        unicodeRanges = font.info.openTypeOS2UnicodeRanges
        self.unicodeRangesEdit = QListView(self)  # too long for BitListView
        self.unicodeRangesEdit.setMinimumHeight(200)
        self.unicodeRangesEdit.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        uRModel = QStandardItemModel(len(unicodeRangesItems), 1)
        for index, elem in enumerate(unicodeRangesItems):
            item = QStandardItem()
            item.setText(elem)
            item.setCheckable(True)
            bit = index
            if unicodeRanges is not None and bit in unicodeRanges:
                item.setCheckState(Qt.Checked)
            else:
                item.setCheckState(Qt.Unchecked)
            uRModel.setItem(index, item)
        self.unicodeRangesEdit.setModel(uRModel)
        unicodeRangesLabel.clicked.connect(self.unicodeRangesEdit.setEnabled)
        if unicodeRanges is not None:
            unicodeRangesLabel.setChecked(True)
        else:
            self.unicodeRangesEdit.setEnabled(False)
        os2AttributesLayout.addRow(unicodeRangesLabel, self.unicodeRangesEdit)

        codePageRangesLabel = RCheckBox(self.tr("Code page ranges:"))
        codePageRangesItems = [
            self.tr("Latin 1 (1252)"),
            self.tr("Latin 2: Eastern Europe (1250)"),
            self.tr("Cyrillic (1251)"),
            self.tr("Greek (1253)"),
            self.tr("Turkish (1254)"),
            self.tr("Hebrew (1255)"),
            self.tr("Arabic (1256)"),
            self.tr("Windows Baltic (1257)"),
            self.tr("Vietnamese (1258)"),
            self.tr("Thai (874)"),
            self.tr("JIS/Japan (932)"),
            self.tr("Chinese: Simplified chars, PRC and Singapore (936)"),
            self.tr("Korean Wansung (949)"),
            self.tr("Chinese: Traditional chars, Taiwan and Hong Kong (950)"),
            self.tr("Korean Johab (1361)"),
            self.tr("Macintosh Character Set (US Roman)"),
            self.tr("OEM Character Set"),
            self.tr("Symbol Character Set"),
            self.tr("IBM Greek (869)"),
            self.tr("MS-DOS Russian (866)"),
            self.tr("MS-DOS Nordic (865)"),
            self.tr("Arabic (864)"),
            self.tr("MS-DOS Canadian French (863)"),
            self.tr("Hebrew (862)"),
            self.tr("MS-DOS Icelandic (861)"),
            self.tr("MS-DOS Portuguese (860)"),
            self.tr("IBM Turkish (857)"),
            self.tr("IBM Cyrillic; primarily Russian (855)"),
            self.tr("Latin 2 (852)"),
            self.tr("MS-DOS Baltic (775)"),
            self.tr("Greek; former 437 G (737)"),
            self.tr("Arabic; ASMO 708 (708)"),
            self.tr("WE/Latin 1 (850)"),
            self.tr("US (437)"),
        ]
        codePageRanges = font.info.openTypeOS2CodePageRanges
        self.codePageRangesEdit = QListView(self)  # too long for BitListView
        self.codePageRangesEdit.setMinimumHeight(200)
        self.codePageRangesEdit.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        cPRModel = QStandardItemModel(len(codePageRangesItems), 1)
        for index, elem in enumerate(codePageRangesItems):
            item = QStandardItem()
            item.setText(elem)
            item.setCheckable(True)
            bit = index
            if index > 17:
                bit += 30
            elif index > 14:
                bit += 14
            elif index > 8:
                bit += 7
            if codePageRanges is not None and bit in codePageRanges:
                item.setCheckState(Qt.Checked)
            else:
                item.setCheckState(Qt.Unchecked)
            cPRModel.setItem(index, item)
        self.codePageRangesEdit.setModel(cPRModel)
        codePageRangesLabel.clicked.connect(self.codePageRangesEdit.setEnabled)
        if codePageRanges is not None:
            codePageRangesLabel.setChecked(True)
        else:
            self.codePageRangesEdit.setEnabled(False)
        os2AttributesLayout.addRow(codePageRangesLabel, self.codePageRangesEdit)

        self.loadCustomIntoForm(
            "openTypeOS2TypoAscender",
            "i",
            self.tr("TypoAscender:"),
            os2AttributesLayout,
        )
        self.loadCustomIntoForm(
            "openTypeOS2TypoDescender",
            "i",
            self.tr("TypoDescender:"),
            os2AttributesLayout,
        )
        self.loadCustomIntoForm(
            "openTypeOS2TypoLineGap", "i", self.tr("TypoLineGap:"), os2AttributesLayout
        )
        self.loadCustomIntoForm(
            "openTypeOS2WinAscent", "pi", self.tr("usWinAscent:"), os2AttributesLayout
        )
        self.loadCustomIntoForm(
            "openTypeOS2WinDescent", "pi", self.tr("usWinDescent:"), os2AttributesLayout
        )
        self.loadCustomIntoForm(
            "openTypeOS2SubscriptXSize",
            "i",
            self.tr("ySubscriptXSize:"),
            os2AttributesLayout,
        )
        self.loadCustomIntoForm(
            "openTypeOS2SubscriptYSize",
            "i",
            self.tr("ySubscriptYSize:"),
            os2AttributesLayout,
        )
        self.loadCustomIntoForm(
            "openTypeOS2SubscriptXOffset",
            "i",
            self.tr("ySubscriptXOffset:"),
            os2AttributesLayout,
        )
        self.loadCustomIntoForm(
            "openTypeOS2SubscriptYOffset",
            "i",
            self.tr("ySubscriptYOffset:"),
            os2AttributesLayout,
        )
        self.loadCustomIntoForm(
            "openTypeOS2SuperscriptXSize",
            "i",
            self.tr("ySuperscriptXSize:"),
            os2AttributesLayout,
        )
        self.loadCustomIntoForm(
            "openTypeOS2SuperscriptYSize",
            "i",
            self.tr("ySuperscriptYSize:"),
            os2AttributesLayout,
        )
        self.loadCustomIntoForm(
            "openTypeOS2SuperscriptXOffset",
            "i",
            self.tr("ySuperscriptXOffset:"),
            os2AttributesLayout,
        )
        self.loadCustomIntoForm(
            "openTypeOS2SuperscriptYOffset",
            "i",
            self.tr("ySuperscriptYOffset:"),
            os2AttributesLayout,
        )
        self.loadCustomIntoForm(
            "openTypeOS2StrikeoutSize",
            "i",
            self.tr("yStrikeoutSize:"),
            os2AttributesLayout,
        )
        self.loadCustomIntoForm(
            "openTypeOS2StrikeoutPosition",
            "i",
            self.tr("yStrikeoutPosition:"),
            os2AttributesLayout,
        )

        os2Layout.addLayout(os2AttributesLayout)

        return os2Group

    def _updateFsTypeVisibility(self, index):
        if index == 0 or not self.fsTypeDrop.isEnabled():
            self.noSubsettingBox.setEnabled(False)
            self.bitmapEmbeddingOnlyBox.setEnabled(False)
        else:
            self.noSubsettingBox.setEnabled(True)
            self.bitmapEmbeddingOnlyBox.setEnabled(True)

    def storeValues(self):
        font = self.font

        # head table
        self.store("openTypeHeadLowestRecPPEM")
        if self.flagsEdit.isEnabled():
            headFlags = []
            itemToFlagMap = {0: 0, 1: 1, 2: 2, 3: 3, 4: 4, 5: 11, 6: 12, 7: 13, 8: 14}
            model = self.flagsEdit.model()
            for index in range(model.rowCount()):
                item = model.item(index)
                if item.checkState() == Qt.Checked:
                    headFlags.append(itemToFlagMap[index])
            font.info.openTypeHeadFlags = headFlags
        else:
            font.info.openTypeHeadFlags = None

        # name table
        self.store("openTypeNamePreferredFamilyName")
        self.store("openTypeNamePreferredSubfamilyName")
        self.store("openTypeNameCompatibleFullName")
        self.store("openTypeNameWWSFamilyName")
        self.store("openTypeNameWWSSubfamilyName")
        # XXX: Implement name records
        self.store("openTypeNameVersion")
        self.store("openTypeNameUniqueID")
        self.store("openTypeNameDescription")
        self.store("openTypeNameSampleText")

        # hhea table
        self.store("openTypeHheaAscender")
        self.store("openTypeHheaDescender")
        self.store("openTypeHheaLineGap")
        self.store("openTypeHheaCaretSlopeRise")
        self.store("openTypeHheaCaretSlopeRun")
        self.store("openTypeHheaCaretOffset")

        # vhea table
        self.store("openTypeVheaVertTypoAscender")
        self.store("openTypeVheaVertTypoDescender")
        self.store("openTypeVheaVertTypoLineGap")
        self.store("openTypeVheaCaretSlopeRise")
        self.store("openTypeVheaCaretSlopeRun")
        self.store("openTypeVheaCaretOffset")

        # XXX: Implement GASP table

        # TODO: see if data needs to be padded to 4 chars.
        # I think that this is to be deferred to ufo2fdk(?)
        self.store("openTypeOS2VendorID")

        if self.fsSelectionList.isEnabled():
            itemToFlagMap = {0: 1, 1: 2, 2: 3, 3: 4, 4: 7, 5: 8, 6: 9}
            fsSelectionModel = self.fsSelectionList.model()
            fsSelection = []
            for i in range(7):
                item = fsSelectionModel.item(i)
                if item.checkState() == Qt.Checked:
                    fsSelection.append(itemToFlagMap[i])
            font.info.openTypeOS2Selection = fsSelection
        else:
            font.info.openTypeOS2Selection = None

        if self.fsTypeDrop.isEnabled():
            fsTypeIndex = self.fsTypeDrop.currentIndex()
            fsType = []
            if fsTypeIndex > 0:
                fsType.append(fsTypeIndex)
                if self.noSubsettingBox.isChecked():
                    fsType.append(8)
                if self.bitmapEmbeddingOnlyBox.isChecked():
                    fsType.append(9)
            font.info.openTypeOS2Type = fsType
        else:
            font.info.openTypeOS2Type = None

        if self.usWidthClassDrop.isEnabled():
            font.info.openTypeOS2WidthClass = self.usWidthClassDrop.currentIndex() + 1
        else:
            font.info.openTypeOS2WidthClass = None

        if self.usWeightClassDrop.isEnabled():
            font.info.openTypeOS2WeightClass = (
                self.usWeightClassDrop.currentIndex() + 1
            ) * 100
        else:
            font.info.openTypeOS2WeightClass = None

        if self.panoseFamilyDrop.isEnabled():
            panose = [
                self.panoseFamilyDrop.currentIndex(),
                self.panoseSerifsDrop.currentIndex(),
                self.panoseWeightDrop.currentIndex(),
                self.panoseProportionDrop.currentIndex(),
                self.panoseContrastDrop.currentIndex(),
                self.panoseStrokeVariationDrop.currentIndex(),
                self.panoseArmStyleDrop.currentIndex(),
                self.panoseLetterformDrop.currentIndex(),
                self.panoseMidlineDrop.currentIndex(),
                self.panoseXHeightDrop.currentIndex(),
            ]
            font.info.openTypeOS2Panose = panose
        else:
            font.info.openTypeOS2Panose = None

        if self.unicodeRangesEdit.isEnabled():
            unicodeRanges = []
            model = self.unicodeRangesEdit.model()
            for index in range(model.rowCount()):
                item = model.item(index)
                if item.checkState() == Qt.Checked:
                    unicodeRanges.append(index)
            font.info.openTypeOS2UnicodeRanges = unicodeRanges
        else:
            font.info.openTypeOS2UnicodeRanges = None

        if self.codePageRangesEdit.isEnabled():
            codePageRanges = []
            model = self.codePageRangesEdit.model()
            for index in range(model.rowCount()):
                item = model.item(index)
                if item.checkState() == Qt.Checked:
                    bit = index
                    if index > 17:
                        bit += 30
                    elif index > 14:
                        bit += 14
                    elif index > 8:
                        bit += 7
                    codePageRanges.append(bit)
            font.info.openTypeOS2CodePageRanges = codePageRanges
        else:
            font.info.openTypeOS2CodePageRanges = None

        self.store("openTypeOS2TypoAscender")
        self.store("openTypeOS2TypoDescender")
        self.store("openTypeOS2TypoLineGap")
        self.store("openTypeOS2WinAscent")
        self.store("openTypeOS2WinDescent")
        self.store("openTypeOS2SubscriptXSize")
        self.store("openTypeOS2SubscriptYSize")
        self.store("openTypeOS2SubscriptXOffset")
        self.store("openTypeOS2SubscriptYOffset")
        self.store("openTypeOS2SuperscriptXSize")
        self.store("openTypeOS2SuperscriptYSize")
        self.store("openTypeOS2SuperscriptXOffset")
        self.store("openTypeOS2SuperscriptYOffset")
        self.store("openTypeOS2StrikeoutSize")
        self.store("openTypeOS2StrikeoutPosition")


class PostScriptTab(TabWidget):
    def __init__(self, font, parent=None):
        super().__init__(font, parent)
        self.name = self.tr("PostScript")

        mainLayout = QVBoxLayout(self)
        mainLayout.setContentsMargins(0, 0, 0, 0)

        tableScrollArea = TransparentScrollArea(self)
        tableScrollArea.setWidgetResizable(True)
        tableArea = QWidget(tableScrollArea)
        tableScrollArea.setWidget(tableArea)
        mainLayout.addWidget(tableScrollArea)

        tableLayout = QVBoxLayout(tableArea)

        # Naming
        namingGroup = QGroupBox(self.tr("Naming"))
        namingLayout = QFormLayout(namingGroup)
        namingLayout.setLabelAlignment(
            Qt.AlignRight | Qt.AlignTrailing | Qt.AlignVCenter
        )

        self.loadCustomIntoForm(
            "postscriptFontName", "s", self.tr("FontName:"), namingLayout
        )
        self.loadCustomIntoForm(
            "postscriptFullName", "s", self.tr("FullName:"), namingLayout
        )
        self.loadCustomIntoForm(
            "postscriptWeightName", "s", self.tr("WeightName:"), namingLayout
        )
        self.loadCustomIntoForm(
            "postscriptUniqueID", "i", self.tr("UniqueID:"), namingLayout
        )

        tableLayout.addWidget(namingGroup)

        # Hints
        hintsGroup = QGroupBox(self.tr("Hints"))
        hintsLayout = QFormLayout(hintsGroup)
        hintsLayout.setLabelAlignment(
            Qt.AlignRight | Qt.AlignTrailing | Qt.AlignVCenter
        )

        self.loadCustomIntoForm(
            "postscriptBlueValues", "if+", self.tr("BlueValues:"), hintsLayout
        )
        self.loadCustomIntoForm(
            "postscriptOtherBlues", "if+", self.tr("OtherBlues:"), hintsLayout
        )
        self.loadCustomIntoForm(
            "postscriptFamilyBlues", "if+", self.tr("FamilyBlues:"), hintsLayout
        )
        self.loadCustomIntoForm(
            "postscriptFamilyOtherBlues",
            "if+",
            self.tr("FamilyOtherBlues:"),
            hintsLayout,
        )

        self.loadCustomIntoForm(
            "postscriptBlueFuzz", "if", self.tr("BlueFuzz:"), hintsLayout
        )
        self.loadCustomIntoForm(
            "postscriptBlueScale", "if", self.tr("BlueScale:"), hintsLayout
        )
        self.loadCustomIntoForm(
            "postscriptBlueShift", "if", self.tr("BlueShift:"), hintsLayout
        )

        self.loadCustomIntoForm(
            "postscriptStemSnapH", "if+", self.tr("StemSnapH:"), hintsLayout
        )
        self.loadCustomIntoForm(
            "postscriptStemSnapV", "if+", self.tr("StemSnapV:"), hintsLayout
        )

        self.loadCustomIntoForm(
            "postscriptForceBold", "b", self.tr("ForceBold:"), hintsLayout
        )

        tableLayout.addWidget(hintsGroup)

        # Metrics
        metricsGroup = QGroupBox(self.tr("Metrics"))
        metricsLayout = QFormLayout(metricsGroup)
        metricsLayout.setLabelAlignment(
            Qt.AlignRight | Qt.AlignTrailing | Qt.AlignVCenter
        )

        self.loadCustomIntoForm(
            "postscriptDefaultWidthX", "if", self.tr("DefaultWidthX:"), metricsLayout
        )
        self.loadCustomIntoForm(
            "postscriptNominalWidthX", "if", self.tr("NominalWidthX:"), metricsLayout
        )
        self.loadCustomIntoForm(
            "postscriptUnderlineThickness",
            "if",
            self.tr("UnderlineThickness:"),
            metricsLayout,
        )
        self.loadCustomIntoForm(
            "postscriptUnderlinePosition",
            "if",
            self.tr("UnderlinePosition:"),
            metricsLayout,
        )
        self.loadCustomIntoForm(
            "postscriptSlantAngle", "if", self.tr("SlantAngle:"), metricsLayout
        )
        self.loadCustomIntoForm(
            "postscriptIsFixedPitch", "b", self.tr("isFixedPitch:"), metricsLayout
        )

        tableLayout.addWidget(metricsGroup)

        # Characters
        charactersGroup = QGroupBox(self.tr("Characters"))
        charactersLayout = QFormLayout(charactersGroup)
        charactersLayout.setLabelAlignment(
            Qt.AlignRight | Qt.AlignTrailing | Qt.AlignVCenter
        )

        self.loadCustomIntoForm(
            "postscriptDefaultCharacter",
            "s",
            self.tr("Default character:"),
            charactersLayout,
        )

        windowsCharacterSetLabel = RCheckBox(self.tr("Windows character set:"))
        self.windowsCharacterSetDrop = QComboBox(self)
        windowsCharacterSetLabel.clicked.connect(
            self.windowsCharacterSetDrop.setEnabled
        )
        items = [
            self.tr("ANSI"),
            self.tr("Default"),
            self.tr("Symbol"),
            self.tr("Macintosh"),
            self.tr("Shift JIS"),
            self.tr("Hangul"),
            self.tr("Hangul (Johab)"),
            self.tr("GB2312"),
            self.tr("Chinese BIG5"),
            self.tr("Greek"),
            self.tr("Turkish"),
            self.tr("Vietnamese"),
            self.tr("Hebrew"),
            self.tr("Arabic"),
            self.tr("Baltic"),
            self.tr("Bitstream"),
            self.tr("Cyrillic"),
            self.tr("Thai"),
            self.tr("Eastern European"),
            self.tr("OEM"),
        ]
        self.windowsCharacterSetDrop.insertItems(0, items)
        if font.info.postscriptWindowsCharacterSet is not None:
            windowsCharacterSetLabel.setChecked(True)
            self.windowsCharacterSetDrop.setCurrentIndex(
                font.info.postscriptWindowsCharacterSet
            )
        else:
            self.windowsCharacterSetDrop.setEnabled(False)
        charactersLayout.addRow(windowsCharacterSetLabel, self.windowsCharacterSetDrop)

        tableLayout.addWidget(charactersGroup)

        spacer = QSpacerItem(10, 20, QSizePolicy.Minimum, QSizePolicy.Expanding)
        tableLayout.addItem(spacer)

    def storeValues(self):
        self.store("postscriptFontName")
        self.store("postscriptFullName")
        self.store("postscriptWeightName")
        self.store("postscriptUniqueID")

        self.store("postscriptBlueValues")
        self.store("postscriptOtherBlues")
        self.store("postscriptFamilyBlues")
        self.store("postscriptFamilyOtherBlues")
        self.store("postscriptBlueFuzz")
        self.store("postscriptBlueScale")
        self.store("postscriptBlueShift")
        self.store("postscriptStemSnapH")
        self.store("postscriptStemSnapV")

        self.store("postscriptDefaultWidthX")
        self.store("postscriptNominalWidthX")
        self.store("postscriptUnderlineThickness")
        self.store("postscriptUnderlinePosition")
        self.store("postscriptSlantAngle")
        self.store("postscriptForceBold")
        self.store("postscriptIsFixedPitch")

        self.store("postscriptDefaultCharacter")
        font = self.font
        windowsCharacterSet = self.windowsCharacterSetDrop
        if windowsCharacterSet.isEnabled():
            font.info.postscriptWindowsCharacterSet = windowsCharacterSet.currentIndex()
        else:
            font.info.postscriptWindowsCharacterSet = None


# ------------------
# Supporting widgets
# ------------------


class BitListView(QListView):
    """
    A QListView to display bit fields. Resizes automatically to fit all flags
    without vertical scroll bars, no horizontal scroll bars.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)

    # https://forum.qt.io/topic/40717/set-size-of-the-qlistview-to-fit-to-it-s-content/7  # noqa
    def sizeHint(self):
        hint = super().sizeHint()

        model = self.model()
        if model:
            extraHeight = self.height() - self.viewport().height()
            vRect = self.visualRect(
                model.index(model.rowCount() - 1, self.modelColumn())
            )
            hint.setHeight(vRect.y() + vRect.height() + extraHeight)

        return hint


class FloatSpinBox(QDoubleSpinBox):
    """
    Hides decimals unless used but shows decimal point to hint at the type of
    box. No more "250,00000000", just "250,".
    """

    def textFromValue(self, value):
        decimalPoint = QLocale().decimalPoint()
        if value.is_integer():
            return "{}{}".format(str(int(value)), decimalPoint)
        else:
            return str(value).replace(".", decimalPoint)


class RCheckBox(QCheckBox):
    """
    A QCheckBox with reversed check box for better alignment of custom
    parameter layouts.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setLayoutDirection(Qt.RightToLeft)



from PyQt5.QtCore import QSize, Qt
from PyQt5.QtGui import QColor, QPainter, QPainterPath
from PyQt5.QtWidgets import QHBoxLayout, QLabel, QSizePolicy, QSpinBox, QWidget

from trufont.controls.pathButton import PathButton

__all__ = ["StatusBar"]

_minPath = QPainterPath()
_minPath.moveTo(1, 6)
_minPath.lineTo(11, 6)

_plusPath = QPainterPath(_minPath)
_plusPath.moveTo(6, 1)
_plusPath.lineTo(6, 11)

_minPath.translate(5, 7)
_plusPath.translate(5, 7)


def Button():
    btn = PathButton()
    btn.setIsDownColor(QColor(210, 210, 210))
    btn.setIsFlipped(True)
    btn.setSize(QSize(23, 25))
    return btn


class StatusBar(QWidget):
    """
    Use the *sizeChanged* signal for size changes.

    TODO: specify only isFontTab/isGlyphTab and put the details
    in the widget internals
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self._shouldPropagateSize = True

        self.statusLabel = QLabel(self)

        btnColor = QColor(126, 126, 126)
        minusButton = Button()
        minusButton.setDrawingCommands([QSize(23, 25), (_minPath, "1", btnColor)])
        minusButton.setProperty("delta", -10)
        minusButton.pressed.connect(self._sizeOffset)
        self.sizeEdit = QSpinBox(self)
        self.sizeEdit.setButtonSymbols(QSpinBox.NoButtons)
        self.sizeEdit.setFixedWidth(56)
        self.sizeEdit.setFrame(False)
        self.sizeEdit.lineEdit().setAlignment(Qt.AlignCenter)
        plusButton = Button()
        plusButton.setDrawingCommands([QSize(23, 25), (_plusPath, "1", btnColor)])
        plusButton.setProperty("delta", 10)
        plusButton.pressed.connect(self._sizeOffset)

        layout = QHBoxLayout(self)
        layout.addWidget(self.statusLabel)
        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        layout.addWidget(spacer)
        layout.addWidget(minusButton)
        layout.addWidget(self.sizeEdit)
        layout.addWidget(plusButton)
        layout.setContentsMargins(15, 0, 15, 0)
        layout.setSpacing(0)

        self.sizeChanged = self.sizeEdit.valueChanged

    def text(self):
        return self.statusLabel.text()

    def setText(self, text):
        self.statusLabel.setText(text)

    def textVisible(self):
        return self.statusLabel.isVisible()

    def setTextVisible(self, value):
        self.statusLabel.setVisible(value)

    def minimumSize(self):
        return self.sizeEdit.minimum()

    def setMinimumSize(self, value):
        self.sizeEdit.blockSignals(True)
        self.sizeEdit.setMinimum(value)
        self.sizeEdit.blockSignals(False)

    def maximumSize(self):
        return self.sizeEdit.maximum()

    def setMaximumSize(self, value):
        self.sizeEdit.blockSignals(True)
        self.sizeEdit.setMaximum(value)
        self.sizeEdit.blockSignals(False)

    def shouldPropagateSize(self):
        return self._shouldPropagateSize

    def setShouldPropagateSize(self, value):
        self._shouldPropagateSize = value

    def unit(self):
        return self.sizeEdit.suffix()

    def setUnit(self, txt):
        self.sizeEdit.setSuffix(txt)

    def size(self):
        return self.sizeEdit.value()

    def setSize(self, value):
        value = round(value)
        self.sizeEdit.blockSignals(True)
        self.sizeEdit.setValue(value)
        self.sizeEdit.blockSignals(False)
        if self._shouldPropagateSize:
            self.sizeEdit.valueChanged.emit(value)
        # nudge label w unclamped value
        self._sliderSizeChanged(value)

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.fillRect(event.rect(), Qt.white)

    def _sizeOffset(self):
        delta = self.sender().property("delta")
        cellsize = self.sizeEdit.value()
        newValue = cellsize + delta
        self.sizeEdit.setValue(newValue)

    def _sliderSizeChanged(self, value):
        self.sizeEdit.setValue(value)



from PyQt5.QtCore import QPointF, QRectF, Qt
from PyQt5.QtGui import QPainterPath
from PyQt5.QtWidgets import QApplication

from trufont.drawingTools.baseTool import BaseTool
from trufont.tools.uiMethods import moveUIPoint

_path = QPainterPath()
_path.moveTo(6.05, 22.5)
_path.lineTo(11.84, 15.59)
_path.cubicTo(11.65, 15.31, 11.47, 14.82, 11.47, 14.34)
_path.cubicTo(11.47, 12.92, 12.58, 11.9, 13.85, 11.9)
_path.cubicTo(15.21, 11.91, 16.27, 12.95, 16.27, 14.34)
_path.cubicTo(16.27, 15.58, 15.26, 16.72, 13.87, 16.72)
_path.cubicTo(13.57, 16.72, 13.19, 16.65, 12.94, 16.53)
_path.lineTo(7.16, 23.43)
_path.lineTo(7.74, 23.92)
_path.cubicTo(7.74, 23.92, 17.62, 20.39, 19.35, 18.33)
_path.cubicTo(21.25, 16.06, 21.37, 13.19, 19.67, 10.61)
_path.lineTo(22.85, 6.83)
_path.lineTo(19.67, 4.17)
_path.lineTo(16.5, 7.95)
_path.cubicTo(13.66, 6.72, 10.84, 7.32, 8.92, 9.58)
_path.cubicTo(7.27, 11.52, 5.47, 22.01, 5.47, 22.01)
_path.closeSubpath()


class PenTool(BaseTool):
    icon = _path
    name = QApplication.translate("PenTool", "Pen")
    shortcut = "P"

    def __init__(self, parent=None):
        super().__init__(parent)
        self._shouldMoveOnCurve = False
        self._stashedOffCurve = None
        self._targetContour = None

    def toolActivated(self):
        glyph = self._glyph
        glyph.addObserver(self, "_selectionChanged", "Glyph.SelectionChanged")

    def toolDisabled(self):
        glyph = self._glyph
        glyph.removeObserver(self, "Glyph.SelectionChanged")
        self._cleanupTrailingOffcurve()
        assert not self._shouldMoveOnCurve

    # notifications

    def _selectionChanged(self, notification):
        if len(self._glyph.selection) != 1:
            self._cleanupTrailingOffcurve()

    # helpers

    def _cleanupTrailingOffcurve(self):
        for contour in self._glyph:
            if not (contour and contour.open):
                continue
            point = contour[-1]
            if point.segmentType is None:
                # TODO(meta): should we emit selection changed if deleted point
                # was selected?
                contour.removePoint(point)
                contour[-1].smooth = False

    def _getSelectedCandidatePoint(self):
        """
        If there is exactly one point selected in the glyph at the edge of an
        open contour, return said contour.

        Else return None.
        """

        candidates = set()
        shouldReverse = False
        for contour in self._glyph:
            # we're after an open contour...
            if not (len(contour) and contour.open):
                continue
            # ...with the last point selected
            sel = contour.selection
            if not len(sel) == 1:
                continue
            lastPoint = contour[-1]
            if not lastPoint.selected:
                if not contour[0].selected:
                    continue
                shouldReverse = True
            else:
                shouldReverse = False
            candidates.add(contour)
        if len(candidates) == 1:
            ret = next(iter(candidates))
            if shouldReverse:
                ret.reverse()
            return ret
        return None

    def _coerceSegmentToCurve(self, contour, pt, pos):
        contour.holdNotifications()
        index = contour.index(pt) or len(contour)
        otherPt = contour.getPoint(index - 1)
        # add an offCurve before pt
        if contour.open:
            # inverse point
            secondX = 2 * pt.x - pos.x()
            secondY = 2 * pt.y - pos.y()
            smooth = True
        else:
            # closed contour we pull the point with the mouse
            secondX = pos.x()
            secondY = pos.y()
            smooth = False
        contour.insertPoint(index, pt.__class__((secondX, secondY)))
        # add the first of two offCurves
        if self._stashedOffCurve is not None:
            offCurve, onSmooth = self._stashedOffCurve
            otherPt.smooth = index - 1 and onSmooth
            contour.insertPoint(index, offCurve)
            self._stashedOffCurve = None
        else:
            firstX = otherPt.x + round(0.35 * (pt.x - otherPt.x))
            firstY = otherPt.y + round(0.35 * (pt.y - otherPt.y))
            contour.insertPoint(index, pt.__class__((firstX, firstY)))
        # now flag pt as curve point
        pt.segmentType = "curve"
        pt.smooth = smooth
        contour.releaseHeldNotifications()

    def _updateOnCurveSmoothness(self, event):
        contour = self._targetContour
        if contour is not None:
            if len(contour) < 2:
                return
            pt = contour[-1]
            if pt.selected:
                # grab the onCurve. if not the previous point, it's the one
                # before it
                if pt.segmentType is None:
                    pt = contour[-2]
                    # this shouldn't happen, but guard whatsoever
                    if pt.segmentType is None:
                        return
                    if pt == contour[0]:
                        return
                pt.smooth = not event.modifiers() & Qt.AltModifier
                contour.dirty = True

    # events

    def keyPressEvent(self, event):
        self._updateOnCurveSmoothness(event)
        self._shouldMoveOnCurve = event.key() == Qt.Key_Space

    def keyReleaseEvent(self, event):
        self._updateOnCurveSmoothness(event)
        if event.key() == Qt.Key_Space:
            self._shouldMoveOnCurve = False

    def mousePressEvent(self, event):
        if event.button() != Qt.LeftButton:
            super().mousePressEvent(event)
            return
        self._glyph.beginUndoGroup()
        self._origin = event.localPos()
        widget = self.parent()
        candidate = self._getSelectedCandidatePoint()
        mouseItem = widget.itemAt(self._origin)
        # if we clicked on an item, see if we should join the current contour
        if isinstance(mouseItem, tuple):
            contour, index = mouseItem
            if contour == candidate and not index:
                point = contour[index]
                lastPoint = candidate[-1]
                lastPoint.selected = False
                if lastPoint.segmentType is not None:
                    point.segmentType = "line"
                else:
                    contour.removePoint(lastPoint)
                    self._stashedOffCurve = (lastPoint, contour[-1].smooth)
                    contour[-1].smooth = False
                point.segmentType = "line"
                point.selected = True
                point.smooth = False
                candidate.postNotification(notification="Contour.SelectionChanged")
                candidate.dirty = True
                self._targetContour = candidate
                return
        canvasPos = event.pos()
        x, y = canvasPos.x(), canvasPos.y()
        # otherwise, add a point to current contour if applicable
        if candidate is not None:
            contour = candidate
            lastPoint = contour[-1]
            lastPoint.selected = False
            if event.modifiers() & Qt.ShiftModifier:
                pos = self.clampToOrigin(
                    self._origin, QPointF(lastPoint.x, lastPoint.y)
                ).toPoint()
                x, y = pos.x(), pos.y()
            if not lastPoint.segmentType:
                contour.removePoint(lastPoint)
                self._stashedOffCurve = (lastPoint, contour[-1].smooth)
                contour[-1].smooth = False
            pointType = "line"
        # or create a new one
        else:
            contour = self._glyph.instantiateContour()
            self._glyph.appendContour(contour)
            pointType = "move"
        # in any case here, unselect all points (*click*) and enable new point
        self._glyph.selected = False
        contour.addPoint((x, y), pointType)
        contour[-1].selected = True
        contour.postNotification(notification="Contour.SelectionChanged")
        self._targetContour = contour

    def mouseMoveEvent(self, event):
        if not event.buttons() & Qt.LeftButton:
            super().mouseMoveEvent(event)
            return
        contour = self._targetContour
        if contour is None:
            return
        # we don't make any check here, mousePressEvent did it for us
        pos = event.pos()
        # selected point
        pt = contour[-1]
        if not contour.open:
            pt_ = contour[0]
            if pt_.selected:
                pt = pt_
        if pt.segmentType and not self._shouldMoveOnCurve:
            # don't make a curve until enough distance is reached
            widget = self.parent()
            rect = QRectF(self._origin, event.localPos())
            widgetRect = widget.mapRectFromCanvas(rect)
            if (widgetRect.bottomRight() - widgetRect.topLeft()).manhattanLength() < 10:
                return
            # go
            onSmooth = not event.modifiers() & Qt.AltModifier
            pt.selected = False
            pt.smooth = len(contour) > 1 and onSmooth
            contour.holdNotifications()
            if pt.segmentType == "line" and onSmooth:
                self._coerceSegmentToCurve(contour, pt, pos)
            elif pt.smooth and contour.open:
                # if there's a curve segment behind, we need to update the
                # offCurve's position to inverse
                if len(contour) > 1:
                    onCurveBefore = contour[-2]
                    onCurveBefore.x = 2 * pt.x - pos.x()
                    onCurveBefore.y = 2 * pt.y - pos.y()
            if contour.open:
                contour.addPoint((pos.x(), pos.y()))
            contour[-1].selected = True
            contour.postNotification(notification="Contour.SelectionChanged")
            contour.releaseHeldNotifications()
        else:
            if pt.segmentType:
                onCurve = pt
            elif contour.open:
                onCurve = contour[-2]
            else:
                onCurve = contour[0]
            if event.modifiers() & Qt.ShiftModifier:
                pos = self.clampToOrigin(
                    event.localPos(), QPointF(onCurve.x, onCurve.y)
                ).toPoint()
            if self._shouldMoveOnCurve:
                dx = pos.x() - pt.x
                dy = pos.y() - pt.y
                moveUIPoint(contour, onCurve, (dx, dy))
            else:
                pt.x = pos.x()
                pt.y = pos.y()
                if contour.open and len(contour) >= 3 and onCurve.smooth:
                    if onCurve.segmentType == "line":
                        self._coerceSegmentToCurve(contour, onCurve, pos)
                    otherSidePoint = contour[-3]
                    otherSidePoint.x = 2 * onCurve.x - pos.x()
                    otherSidePoint.y = 2 * onCurve.y - pos.y()
            contour.dirty = True

    def mouseReleaseEvent(self, event):
        if event.button() == Qt.LeftButton:
            self._shouldMoveOnCurve = False
            self._stashedOffCurve = None
            self._targetContour = None
            self._glyph.endUndoGroup()
        else:
            super().mouseReleaseEvent(event)



from PyQt5.QtCore import QSize
from PyQt5.QtWidgets import QMainWindow, QMessageBox

from defconQt.controls.featureCodeEditor import FeatureCodeEditor
from trufont.controls.fileMessageBoxes import CloseMessageBox
from trufont.objects import settings
from trufont.objects.menu import Entries


class FontFeaturesWindow(QMainWindow):
    def __init__(self, font=None, parent=None):
        super().__init__(parent)

        self.font = font
        self.font.info.addObserver(self, "_fontInfoChanged", "Info.Changed")
        self.editor = FeatureCodeEditor(self)
        self.editor.setPlainText(self.font.features.text)
        self.editor.modificationChanged.connect(self.setWindowModified)

        self.updateWindowTitle()
        self.setCentralWidget(self.editor)

        self.readSettings()

    def readSettings(self):
        geometry = settings.fontFeaturesWindowGeometry()
        if geometry:
            self.restoreGeometry(geometry)

    def writeSettings(self):
        settings.setFontFeaturesWindowGeometry(self.saveGeometry())

    def setupMenu(self, menuBar):
        fileMenu = menuBar.fetchMenu(Entries.File)
        fileMenu.fetchAction(Entries.File_Save, self.saveFile)
        fileMenu.addSeparator()
        fileMenu.fetchAction(Entries.File_Reload, self.reloadFile)
        fileMenu.fetchAction(Entries.File_Close, self.close)

    def updateWindowTitle(self, title=None, font=None):
        if title is None:
            title = self.tr("Font Features")
        if font is None:
            font = self.font
        text = self.tr("{0} – {1} {2}").format(
            title, font.info.familyName, font.info.styleName
        )
        self.setWindowTitle("[*]" + text)

    # -------------
    # Notifications
    # -------------

    def _fontInfoChanged(self, notification):
        self.updateWindowTitle()

    # ------------
    # Menu methods
    # ------------

    def saveFile(self):
        self.editor.write(self.font.features)
        parent = self.parent()
        if parent is not None and hasattr(parent, "saveFile"):
            parent.saveFile()

    def reloadFile(self):
        self.font.reloadFeatures()
        self.editor.setPlainText(self.font.features.text)

    # ----------
    # Qt methods
    # ----------

    def sizeHint(self):
        return QSize(650, 500)

    def moveEvent(self, event):
        self.writeSettings()

    resizeEvent = moveEvent

    def closeEvent(self, event):
        if self.editor.document().isModified():
            ret = CloseMessageBox.getCloseDocument(self)
            if ret == QMessageBox.Save:
                self.saveFile()
                event.accept()
            elif ret == QMessageBox.Discard:
                event.accept()
            else:
                event.ignore()
                return
            self.font.info.removeObserver(self, "Info.Changed")



def characterToGlyphName(c, cmap):
    v = ord(c)
    v = cmap.get(v)
    if isinstance(v, list):
        v = v[0]
    return v


def compileStack(glyphNames, stack):
    if stack:
        glyphNames.append("".join(stack))


def escapeText(text):
    return text.replace("//", "/slash ")


def splitText(
    text,
    cmap,
    fallback=".notdef",
    cmapFunc=characterToGlyphName,
    compileFunc=compileStack,
    escapeFunc=escapeText,
):
    """
    Break a string of characters or / delimited glyph names
    into a list.
    """
    # escape
    text = escapeFunc(text)
    #
    glyphNames = []
    compileStack = None
    for c in text:
        # start a glyph name compile.
        if c == "/":
            # finishing a previous compile.
            if compileStack is not None:
                # only add the compile if something has been added to the stack.
                compileFunc(glyphNames, compileStack)
            # reset the stack.
            compileStack = []
        # adding to or ending a glyph name compile.
        elif compileStack is not None:
            # space. conclude the glyph name compile.
            if c == " ":
                # only add the compile if something has been added to the stack.
                compileFunc(glyphNames, compileStack)
                compileStack = None
            # add the character to the stack.
            else:
                compileStack.append(c)
        # adding a character that needs to be converted to a glyph name.
        else:
            glyphName = cmapFunc(c, cmap)
            if glyphName is None:
                glyphName = fallback
            glyphNames.append(glyphName)
    # catch remaining compile.
    compileFunc(glyphNames, compileStack)
    return glyphNames



from PyQt5.QtCore import QPoint, QRect, QSize, Qt
from PyQt5.QtGui import QIcon, QIconEngine, QPainter, QPixmap


class PathIconEngine(QIconEngine):
    def __init__(self, width=None, height=None):
        super().__init__()
        self._fillPaths = []
        self._strokePaths = []
        if width is None:
            size = QSize()
        else:
            if height is None:
                height = width
            size = QSize(width, height)
        self._size = size

    def addFillPath(self, path, color=Qt.black, antialiasing=False):
        # TODO: add Mode, State
        self._fillPaths.append((path, color, antialiasing))

    def addStrokePath(self, path, color=Qt.black, width=0.9, antialiasing=False):
        # TODO: add Mode, State
        self._strokePaths.append((path, color, width, antialiasing))

    def size(self):
        return self._size

    def setSize(self, width, height=None):
        if height is None:
            height = width
        self._size = (width, height)

    def paint(self, painter, rect, mode, state):
        # TODO: use Mode, State
        if not (self._fillPaths or self._strokePaths):
            return
        painter.save()
        size = self._size
        target = rect.size()
        if not size.isNull() and (
            size.width() > target.width() or size.height() > target.height()
        ):
            sz = size.scaled(target, Qt.KeepAspectRatio)
            width, height = sz.width() / size.width(), sz.height() / size.height()
            # TODO: don't scale the painter, instead make a rect which
            # coordinates are used to paint in; this will ensure pixel
            # perfection
            painter.scale(width, height)
        for path, color, antialiasing in self._fillPaths:
            painter.setRenderHint(QPainter.Antialiasing, antialiasing)
            painter.fillPath(path, color)
        for path, color, width, antialiasing in self._strokePaths:
            painter.setRenderHint(QPainter.Antialiasing, antialiasing)
            pen = painter.pen()
            pen.setColor(color)
            pen.setWidthF(width)
            painter.setPen(pen)
            painter.drawPath(path)
        painter.restore()

    def pixmap(self, size, mode, state):
        pixmap = QPixmap(size)
        pixmap.fill(Qt.transparent)
        painter = QPainter()
        painter.begin(pixmap)
        self.paint(painter, QRect(QPoint(0, 0), size), mode, state)
        painter.end()
        return pixmap


class PathIcon(QIcon):
    def __init__(self, *args):
        self._engine = PathIconEngine(*args)
        super().__init__(self._engine)

    def addFillPath(self, *args, **kwargs):
        self._engine.addFillPath(*args, **kwargs)

    def addStrokePath(self, *args, **kwargs):
        self._engine.addStrokePath(*args, **kwargs)

    def availableSizes(self, mode, state):
        # TODO: use Mode, State
        ret = []
        size = self._engine.size()
        if size.isValid():
            ret.append(size)
        return ret



from PyQt5.QtCore import QLineF, QPointF

from trufont.objects.defcon import TContour
from trufont.tools import bezierMath

# ----------------
# Helper functions
# ----------------


def nudgeUICurve(on1, off1, off2, on2, delta):
    if on2.selected != on1.selected:
        sign = -on1.selected or 1
        sdx, sdy = map(float(sign).__mul__, delta)
        # factor
        xFactor = on2.x - on1.x - sdx
        if xFactor:
            xFactor = (on2.x - on1.x) / xFactor
        yFactor = on2.y - on1.y - sdy
        if yFactor:
            yFactor = (on2.y - on1.y) / yFactor
        # apply
        if not off1.selected:
            off1.x = on1.x + xFactor * (off1.x - on1.x)
            off1.y = on1.y + yFactor * (off1.y - on1.y)
        if not off2.selected:
            off2.x = on1.x + xFactor * (off2.x - on1.x - sdx)
            off2.y = on1.y + yFactor * (off2.y - on1.y - sdy)


def projectUIPointOnRefLine(x1, y1, x2, y2, pt):
    x, y, t = bezierMath.lineProjection(x1, y1, x2, y2, pt.x, pt.y, False)
    # TODO: use grid precision ROUND
    pt.x = x  # round(x)
    pt.y = y  # round(y)


def rotateUIPointAroundRefLine(x1, y1, x2, y2, pt):
    """
    Given three points p1, p2, pt this rotates pt around p2 such that p1,p2 and
    p1,pt are collinear.
    """
    line = QLineF(pt.x, pt.y, x2, y2)
    p2p_l = line.length()
    line.setP1(QPointF(x1, y1))
    p1p2_l = line.length()
    if not p1p2_l:
        return
    line.setLength(p1p2_l + p2p_l)
    # TODO: use grid precision ROUND
    pt.x = line.x2()  # round(line.x2())
    pt.y = line.y2()  # round(line.y2())


# ----------
# Main sauce
# ----------


def UIMove(contour, delta, nudgePoints=False, slidePoints=False):
    contour_ = contour
    # start at an onCurve; situations we want to avoid are:
    # - two offCurves at the start of the contour, since all "handles ops"
    #   (like rotation etc.) need a "next" offCurve after reading the parent
    #   onCurve
    # - one offCurve at the start of the contour, nudge needs to read the
    #   previous onCurve followed by offCurves (followed by onCurve)
    if len(contour) > 2 and contour[0].segmentType is None:
        offset = 1 + (contour[1].segmentType is None)
        contour = contour[offset:] + contour[:offset]
    # first pass: move
    didMove = False
    nextOffShouldMove = False
    nudgeStuff = []
    prev = contour[-1]
    for point in contour:
        if len(nudgeStuff) == 1 and point.segmentType is None:
            nudgeStuff.append(point)
        if point.selected or (nextOffShouldMove and point.segmentType is None):
            point.move(delta)
            didMove = True
        nextOffShouldMove = False
        if point.segmentType and not slidePoints:
            if point.selected:
                # move previous point
                if (
                    prev.segmentType is None
                    and not prev.selected
                    and point.segmentType != "move"
                ):
                    prev.move(delta)
                    didMove = True
                # schedule the next point for move
                nextOffShouldMove = True
            if nudgePoints:
                if len(nudgeStuff) == 2 and point.segmentType == "curve":
                    on1, off1 = nudgeStuff
                    nudgeUICurve(on1, off1, prev, point, delta)
                nudgeStuff = [point]
        prev = point
    if len(nudgeStuff) == 2 and contour[0].segmentType == "curve":
        on1, off1 = nudgeStuff
        nudgeUICurve(on1, off1, prev, contour[0], delta)
    if not didMove:
        return
    del nudgeStuff
    # second pass: constrain
    secondPrevForSliding = None
    secondPrevForRotation = None
    prev = contour[-1]
    for point in contour:
        # slide points
        if secondPrevForSliding is not None and (
            point.segmentType is None or secondPrevForSliding.segmentType is None
        ):
            p1, p2, p3 = secondPrevForSliding, prev, point
            if p2.selected and p2.segmentType != "move":
                if p1.selected or p3.selected:
                    if p1.selected:
                        p1, p3 = p3, p1
                    # if p3 is selected, we just let it move freely
                    if not p1.selected and p3.segmentType is None:
                        rotateUIPointAroundRefLine(p1.x, p1.y, p2.x, p2.y, p3)
                else:
                    if p2.smooth:
                        projectUIPointOnRefLine(p1.x, p1.y, p3.x, p3.y, p2)
            elif p1.selected != p3.selected:
                if p1.selected:
                    p1, p3 = p3, p1
                if p3.segmentType is None:
                    if p2.smooth and p2.segmentType != "move":
                        projectUIPointOnRefLine(p1.x, p1.y, p2.x, p2.y, p3)
                    else:
                        dx, dy = delta
                        rx, ry = p3.x - dx, p3.y - dy
                        projectUIPointOnRefLine(p2.x, p2.y, rx, ry, p3)
        secondPrevForSliding = None
        if slidePoints and point.segmentType is not None:
            secondPrevForSliding = prev
        # rotation/projection across smooth onCurve
        if secondPrevForRotation is not None and (
            point.segmentType is None or secondPrevForRotation.segmentType is None
        ):
            p1, p2, p3 = secondPrevForRotation, prev, point
            if p2.selected:
                if p1.segmentType is None:
                    p1, p3 = p3, p1
                if p1.segmentType is not None:
                    rotateUIPointAroundRefLine(p1.x, p1.y, p2.x, p2.y, p3)
            elif p1.selected != p3.selected:
                if p1.selected:
                    p1, p3 = p3, p1
                if p1.segmentType is not None:
                    projectUIPointOnRefLine(p1.x, p1.y, p2.x, p2.y, p3)
                elif not slidePoints:
                    rotateUIPointAroundRefLine(p3.x, p3.y, p2.x, p2.y, p1)
        secondPrevForRotation = None
        if point.smooth and point.segmentType not in ("move", None):
            secondPrevForRotation = prev
        # --
        prev = point
    contour_.dirty = True


# -----
# Tests
# -----


def UIMove_buildContour(data):
    contour = TContour()
    pen = contour
    pen.beginPath()
    for coords, segmentType, smooth, selected in data:
        pen.addPoint(coords, segmentType, smooth, selected=selected)
    pen.endPath()
    return contour


def UIMove_testContour(contour, data):
    assert len(contour) == len(data), "contour has len {}, expected {}".format(
        len(contour), len(data)
    )
    try:
        for point, (coords, segmentType, smooth, selected) in zip(contour, data):
            # HACK: we round here for now
            point.x = round(point.x)
            point.y = round(point.y)
            assert (point.x, point.y) == coords
            assert point.segmentType == segmentType
            assert point.smooth == smooth
            assert point.selected == selected
    except AssertionError:
        # TODO: pprint and make array from contour instead of doing this
        # manually
        print("contour gives:")
        print("[")
        for point in contour:
            print(
                "    (({}, {}), {}, {}, {}),".format(
                    point.x,
                    point.y,
                    repr(point.segmentType),
                    point.smooth,
                    point.selected,
                )
            )
        print("]")
        print("expected:")
        print("[")
        for line in data:
            print(f"    {line},")
        print("]")
        print()


# runner


def UIMove_runTests():
    UIMove_test_move()
    UIMove_test_move_offWithOn()
    UIMove_test_move_offAtStart()
    UIMove_test_constrain_slidePoints()
    UIMove_test_constrain_smoothOffRotation()
    UIMove_test_constrain_smoothOnProjection()


# tests


def UIMove_test_move():
    """
    Simple test moving onCurve points.

    At closed and open contour boundaries.
    """
    contour = UIMove_buildContour(
        [
            ((0, 0), "move", False, False),
            ((3, 2), "line", False, True),
            ((0, 3), "line", False, False),
        ]
    )
    UIMove(contour, (2, 3))
    UIMove_testContour(
        contour,
        [
            ((0, 0), "move", False, False),
            ((5, 5), "line", False, True),
            ((0, 3), "line", False, False),
        ],
    )
    contour[1].selected = False
    contour[2].selected = True
    UIMove(contour, (2, 3))
    UIMove_testContour(
        contour,
        [
            ((0, 0), "move", False, False),
            ((5, 5), "line", False, False),
            ((2, 6), "line", False, True),
        ],
    )


def UIMove_test_move_offWithOn():
    """
    Moving offCurve points along with their onCurve.

    Make sure offCurve points move with the onCurve, whether or not the
    offCurve is selected.

    Make sure an offCurve moves normally when its onCurve is not selected.

    Edge-case: given an open contour with trailing offCurve, make sure the
    first point moving does not move the trailing offCurve.

    TODO: Edge-case: avoid double move for quadratic curve with single
    offCurve.
    """
    contour = UIMove_buildContour(
        [
            ((3, 0), "move", False, True),
            ((3, 2), None, False, False),
            ((2, 3), None, False, False),
            ((0, 3), "curve", False, False),
            ((-2, 3), None, False, False),
        ]
    )
    UIMove(contour, (1, 2))
    UIMove_testContour(
        contour,
        [
            ((4, 2), "move", False, True),
            ((4, 4), None, False, False),
            ((2, 3), None, False, False),
            ((0, 3), "curve", False, False),
            ((-2, 3), None, False, False),
        ],
    )
    contour[0].selected = False
    contour[2].selected = True
    contour[3].selected = True
    UIMove(contour, (1, 2))
    UIMove_testContour(
        contour,
        [
            ((4, 2), "move", False, False),
            ((4, 4), None, False, False),
            ((3, 5), None, False, True),
            ((1, 5), "curve", False, True),
            ((-1, 5), None, False, False),
        ],
    )
    contour[1].selected = True
    contour[3].smooth = True
    UIMove(contour, (1, 2))
    UIMove_testContour(
        contour,
        [
            ((4, 2), "move", False, False),
            ((5, 6), None, False, True),
            ((4, 7), None, False, True),
            ((2, 7), "curve", True, True),
            ((0, 7), None, False, False),
        ],
    )


def UIMove_test_move_offAtStart():
    """
    Make sure everything works in the special case of two offCurves
    at the beginning of a contour.
    """
    contour = UIMove_buildContour(
        [
            ((2, 2), None, False, True),
            ((1, 3), None, False, True),
            ((0, 5), "curve", False, True),
            ((3, 0), "line", False, True),
        ]
    )
    UIMove(contour, (2, 1))
    UIMove_testContour(
        contour,
        [
            ((4, 3), None, False, True),
            ((3, 4), None, False, True),
            ((2, 6), "curve", False, True),
            ((5, 1), "line", False, True),
        ],
    )


def UIMove_test_constrain_slidePoints():
    """
    Slide points along Bézier handles. Disarm other move relationships
    (offWithOn, onCurve moves offCurve across its sibling smooth on)

    Make sure non-smooth onCurve isn't constrained when moving.
    """
    contour = UIMove_buildContour(
        [
            ((3, 0), "move", False, False),
            ((2, 2), None, False, False),
            ((1, 3), None, False, False),
            ((0, 5), "curve", True, True),
            ((-1, 7), None, False, False),
        ]
    )
    UIMove(contour, (2, 2), slidePoints=True)
    UIMove_testContour(
        contour,
        [
            ((3, 0), "move", False, False),
            ((2, 2), None, False, False),
            ((1, 3), None, False, False),
            ((0, 6), "curve", True, True),
            ((-1, 7), None, False, False),
        ],
    )
    contour[4].selected = True
    UIMove(contour, (2, 0), slidePoints=True)
    UIMove_testContour(
        contour,
        [
            ((3, 0), "move", False, False),
            ((2, 2), None, False, False),
            ((1, 3), None, False, False),
            ((2, 6), "curve", True, True),
            ((2, 7), None, False, True),
        ],
    )
    contour[3].selected = False
    contour[4].selected = False
    contour[1].selected = True
    UIMove(contour, (1, 3), slidePoints=True)
    UIMove_testContour(
        contour,
        [
            ((3, 0), "move", False, False),
            ((1, 4), None, False, True),
            ((1, 3), None, False, False),
            ((2, 6), "curve", True, False),
            ((2, 7), None, False, False),
        ],
    )
    contour[4].segmentType = "line"
    contour[1].selected = False
    contour[4].selected = True
    UIMove(contour, (-2, -3), slidePoints=True)
    UIMove_testContour(
        contour,
        [
            ((3, 0), "move", False, False),
            ((1, 4), None, False, False),
            ((1, 3), None, False, False),
            ((2, 6), "curve", True, False),
            ((0, 4), "line", False, True),
        ],
    )
    contour[4].segmentType = None
    contour[2].selected = True
    contour[3].selected = True
    UIMove(contour, (4, 2), slidePoints=True)
    UIMove_testContour(
        contour,
        [
            ((3, 0), "move", False, False),
            ((1, 4), None, False, False),
            ((5, 5), None, False, True),
            ((6, 8), "curve", True, True),
            ((4, 6), None, False, True),
        ],
    )
    contour[2].selected = False
    contour[4].selected = False
    contour[3].smooth = False
    UIMove(contour, (-4, -2), slidePoints=True)
    UIMove_testContour(
        contour,
        [
            ((3, 0), "move", False, False),
            ((1, 4), None, False, False),
            ((5, 5), None, False, False),
            ((2, 6), "curve", False, True),
            ((4, 6), None, False, False),
        ],
    )


def UIMove_test_constrain_smoothOffRotation():
    """
    Rotate offCurve point across smooth onCurve.
    NOTE: we make sure in the first test to make it fail if the function
    projects instead of rotating (by doing a pi/2 rotation, hence making
    old/new vectors orthogonal).

    Edge-case: don't rotate across open contour boundary; given a trailing
    offCurve.

    Make sure that two offCurve on either side of a smooth point can move
    freely if the smooth point is unselected.
    TODO: is this really the ideal behavior? I guess it makes the most sense.
    """
    contour = UIMove_buildContour(
        [
            ((3, 0), "move", True, False),
            ((3, 2), None, False, False),
            ((2, 4), None, False, True),
            ((0, 3), "curve", True, False),
            ((-2, 2), None, False, False),
        ]
    )
    UIMove(contour, (-3, 1))
    UIMove_testContour(
        contour,
        [
            ((3, 0), "move", True, False),
            ((3, 2), None, False, False),
            ((-1, 5), None, False, True),
            ((0, 3), "curve", True, False),
            ((1, 1), None, False, False),
        ],
    )
    contour[2].selected = False
    contour[1].selected = True
    UIMove(contour, (1, 2))
    UIMove_testContour(
        contour,
        [
            ((3, 0), "move", True, False),
            ((4, 4), None, False, True),
            ((-1, 5), None, False, False),
            ((0, 3), "curve", True, False),
            ((1, 1), None, False, False),
        ],
    )
    contour[1].selected = False
    contour[2].selected = True
    contour[3].selected = True
    UIMove(contour, (1, 2))
    UIMove_testContour(
        contour,
        [
            ((3, 0), "move", True, False),
            ((4, 4), None, False, False),
            ((0, 7), None, False, True),
            ((1, 5), "curve", True, True),
            ((2, 3), None, False, False),
        ],
    )
    contour[3].selected = False
    contour[4].selected = True
    UIMove(contour, (1, 2))
    UIMove_testContour(
        contour,
        [
            ((3, 0), "move", True, False),
            ((4, 4), None, False, False),
            ((1, 9), None, False, True),
            ((1, 5), "curve", True, False),
            ((3, 5), None, False, True),
        ],
    )


def UIMove_test_constrain_smoothOnProjection():
    """
    Project offCurve point across smooth onCurve (with another onCurve on the
    other side).

    Edge-case: don't project across open contour boundary; given a trailing
    offCurve, smooth onCurve first point, and selected onCurve second point.

    Make sure moving the smooth point rotates its sibling offCurve.

    Make two onCurves around a smooth onCurve don't get constrained.

    TODO: handle the case where the onCurve and offCurve on each side of the
    smooth point are selected.
    """
    contour = UIMove_buildContour(
        [
            ((3, 0), "move", True, False),
            ((2, 2), "line", True, False),
            ((1, 4), None, False, True),
            ((0, 5), None, False, False),
            ((-2, 5), "curve", True, False),
            ((-4, 5), "line", True, False),
            ((-6, 5), None, False, False),
        ]
    )
    UIMove(contour, (1, 3))
    UIMove_testContour(
        contour,
        [
            ((3, 0), "move", True, False),
            ((2, 2), "line", True, False),
            ((0, 6), None, False, True),
            ((0, 5), None, False, False),
            ((-2, 5), "curve", True, False),
            ((-4, 5), "line", True, False),
            ((-6, 5), None, False, False),
        ],
    )
    UIMove(contour, (5, -4))
    UIMove_testContour(
        contour,
        [
            ((3, 0), "move", True, False),
            ((2, 2), "line", True, False),
            ((1, 3), None, False, True),
            ((0, 5), None, False, False),
            ((-2, 5), "curve", True, False),
            ((-4, 5), "line", True, False),
            ((-6, 5), None, False, False),
        ],
    )
    contour[2].selected = False
    contour[0].selected = True
    UIMove(contour, (-3, 2))
    UIMove_testContour(
        contour,
        [
            ((0, 2), "move", True, True),
            ((2, 2), "line", True, False),
            ((3, 2), None, False, False),
            ((0, 5), None, False, False),
            ((-2, 5), "curve", True, False),
            ((-4, 5), "line", True, False),
            ((-6, 5), None, False, False),
        ],
    )
    contour[1].selected = True
    UIMove(contour, (1, 2))
    UIMove_testContour(
        contour,
        [
            ((1, 4), "move", True, True),
            ((3, 4), "line", True, True),
            ((4, 4), None, False, False),
            ((0, 5), None, False, False),
            ((-2, 5), "curve", True, False),
            ((-4, 5), "line", True, False),
            ((-6, 5), None, False, False),
        ],
    )
    contour[0].selected = False
    contour[1].selected = False
    contour[5].selected = True
    UIMove(contour, (0, 2))
    UIMove_testContour(
        contour,
        [
            ((1, 4), "move", True, False),
            ((3, 4), "line", True, False),
            ((4, 4), None, False, False),
            ((-1, 4), None, False, False),
            ((-2, 5), "curve", True, False),
            ((-4, 7), "line", True, True),
            ((-5, 8), None, False, False),
        ],
    )
    contour[4].smooth = False
    contour[6].segmentType = "line"
    UIMove(contour, (1, 2))
    UIMove_testContour(
        contour,
        [
            ((1, 4), "move", True, False),
            ((3, 4), "line", True, False),
            ((4, 4), None, False, False),
            ((-1, 4), None, False, False),
            ((-2, 5), "curve", False, False),
            ((-3, 9), "line", True, True),
            ((-5, 8), "line", False, False),
        ],
    )
    contour[5].selected = False
    contour[6].selected = True
    UIMove(contour, (1, 2))
    UIMove_testContour(
        contour,
        [
            ((1, 4), "move", True, False),
            ((3, 4), "line", True, False),
            ((4, 4), None, False, False),
            ((-1, 4), None, False, False),
            ((-2, 5), "curve", False, False),
            ((-3, 9), "line", True, False),
            ((-4, 10), "line", False, True),
        ],
    )



import os

from PyQt5.QtCore import QDir, QFileSystemWatcher, QSize, QStandardPaths, Qt
from PyQt5.QtGui import QColor
from PyQt5.QtWidgets import (
    QCheckBox,
    QDialog,
    QDialogButtonBox,
    QFileDialog,
    QFormLayout,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QStyle,
    QVBoxLayout,
    QWidget,
)
from ufo2ft.fontInfoData import getAttrWithFallback

from defconQt.controls.roundedButtonSet import RoundedButtonSet as ButtonSet
from trufont.objects import icons, settings


class ExportDialog(QDialog):
    def __init__(self, font, parent=None):
        super().__init__(parent, Qt.MSWindowsFixedSizeDialogHint)
        self.setWindowModality(Qt.WindowModal)
        self.setWindowTitle(self.tr("Export…"))

        self._exportDirectory = QDir.toNativeSeparators(
            QStandardPaths.standardLocations(QStandardPaths.DocumentsLocation)[0]
        )
        self.baseName = getAttrWithFallback(font.info, "postscriptFontName")

        self.formatBtnSet = ButtonSet(self)
        self.formatBtnSet.setOptions(["OTF", "TTF"])
        self.formatBtnSet.setSelectionMode(ButtonSet.OneOrMoreSelection)
        self.compressionBtnSet = ButtonSet(self)
        self.compressionBtnSet.setOptions(["None", "WOFF", "WOFF2"])
        self.compressionBtnSet.setSelectionMode(ButtonSet.OneOrMoreSelection)
        self.numberLabel = QLabel(self)
        self.formatBtnSet.clicked.connect(self.updateNumbers)
        self.compressionBtnSet.clicked.connect(self.updateNumbers)

        self.removeOverlapBox = QCheckBox(self.tr("Remove Overlap"), self)
        # self.removeOverlapBox.setChecked(True)  # XXX: implement
        self.removeOverlapBox.setEnabled(False)
        self.autohintBox = QCheckBox(self.tr("Autohint"), self)
        # self.autohintBox.setChecked(True)  # XXX: implement
        self.autohintBox.setEnabled(False)

        self.exportBox = QCheckBox(self)
        boxSize = self.exportBox.sizeHint()
        self.exportBox.setText(self.tr("Use Export Directory"))
        self.exportBox.setChecked(True)
        self.exportIcon = QLabel(self)
        icon = self.style().standardIcon(QStyle.SP_DirClosedIcon)
        iconSize = QSize(24, 24)
        self.exportIcon.setPixmap(icon.pixmap(icon.actualSize(iconSize)))
        self.exportIcon.setBaseSize(iconSize)
        self.exportDirLabel = QLabel(self)
        self.exportDirLabel.setText(self.exportDirectory)
        self.exportDirButton = QPushButton(self)
        self.exportDirButton.setText(self.tr("Choose…"))
        self.exportDirButton.clicked.connect(
            lambda: self.chooseExportDir(self.exportDirectory)
        )

        # if files are to be overwritten, put up a warning
        # + use a file system watcher to avoid TOCTOU
        self.warningIcon = QLabel(self)
        icon = icons.i_warning()
        iconSize = QSize(20, 20)
        self.warningIcon.setPixmap(icon.pixmap(icon.actualSize(iconSize)))
        self.warningIcon.setBaseSize(iconSize)
        # XXX: not sure why this is needed
        self.warningIcon.setFixedWidth(iconSize.width())
        sp = self.warningIcon.sizePolicy()
        sp.setRetainSizeWhenHidden(True)
        self.warningIcon.setSizePolicy(sp)
        self.warningLabel = QLabel(self)
        palette = self.warningLabel.palette()
        role, color = self.warningLabel.foregroundRole(), QColor(230, 20, 20)
        palette.setColor(palette.Active, role, color)
        palette.setColor(palette.Inactive, role, color)
        self.warningLabel.setPalette(palette)
        sp = self.warningLabel.sizePolicy()
        sp.setRetainSizeWhenHidden(True)
        self.warningLabel.setSizePolicy(sp)

        self.updateExportStatus()
        self.exportBox.toggled.connect(self.updateExportStatus)

        self.watcher = QFileSystemWatcher(self)
        self.watcher.addPath(self.exportDirectory)
        self.updateNumbers()
        self.watcher.directoryChanged.connect(self.updateNumbers)

        buttonBox = QDialogButtonBox(QDialogButtonBox.Cancel)
        buttonBox.addButton(self.tr("Generate…"), QDialogButtonBox.AcceptRole)
        buttonBox.accepted.connect(self.finish)
        buttonBox.rejected.connect(self.reject)

        layout = QVBoxLayout(self)
        formLayout = QFormLayout()
        formLayout.addRow(self.tr("Format"), self.formatBtnSet)
        formLayout.addRow(self.tr("Compression"), self.compressionBtnSet)
        formLayout.setHorizontalSpacing(16)
        formLayout.setContentsMargins(0, 0, 0, 4)
        layout.addLayout(formLayout)
        layout.addWidget(self.numberLabel)
        layout.addWidget(self.removeOverlapBox)
        layout.addWidget(self.autohintBox)
        layout.addWidget(self.exportBox)
        exportLayout = QHBoxLayout()
        exportLayout.addWidget(self.exportIcon)
        exportLayout.addWidget(self.exportDirLabel)
        exportLayout.addWidget(self.exportDirButton)
        exportLayout.addWidget(QWidget())
        margins = exportLayout.contentsMargins()
        margins.setLeft(margins.left() + boxSize.width() + 4)
        exportLayout.setContentsMargins(margins)
        exportLayout.setStretch(3, 1)
        layout.addLayout(exportLayout)
        warningLayout = QHBoxLayout()
        warningLayout.addWidget(self.warningIcon)
        warningLayout.addWidget(self.warningLabel)
        warningLayout.addWidget(QWidget())
        margins.setBottom(margins.bottom() + 4)
        warningLayout.setContentsMargins(margins)
        warningLayout.setStretch(3, 1)
        layout.addLayout(warningLayout)
        layout.addWidget(buttonBox)
        # XXX: check this on non-Windows platforms
        layout.setContentsMargins(16, 16, 16, 16)

        self.readSettings()

    def readSettings(self):
        attrs = [
            (settings.exportFileFormats, self.formatBtnSet.setSelectedOptions),
            (
                settings.exportCompressionFormats,
                self.compressionBtnSet.setSelectedOptions,
            ),
            (settings.exportRemoveOverlap, self.removeOverlapBox.setChecked),
            (settings.exportAutohint, self.autohintBox.setChecked),
            (settings.exportUseDirectory, self.exportBox.setChecked),
            (
                settings.exportDirectory,
                lambda attr: setattr(self, "exportDirectory", attr),
            ),
        ]
        for getValue, setter in attrs:
            value = getValue()
            if value != "":
                setter(value)
        self.exportDirLabel.setText(self.exportDirectory)
        self.updateNumbers()

    def writeSettings(self):
        attrs = [
            (settings.setExportFileFormats, self.formatBtnSet.selectedOptions),
            (
                settings.setExportCompressionFormats,
                self.compressionBtnSet.selectedOptions,
            ),
            (settings.setExportRemoveOverlap, self.removeOverlapBox.isChecked),
            (settings.setExportAutohint, self.autohintBox.isChecked),
            (settings.setExportUseDirectory, self.exportBox.isChecked),
            (settings.setExportDirectory, lambda: getattr(self, "exportDirectory")),
        ]
        for setValue, getter in attrs:
            value = getter()
            setValue(value)

    @property
    def exportDirectory(self):
        return self._exportDirectory

    @exportDirectory.setter
    def exportDirectory(self, path):
        oldValue = self._exportDirectory
        if oldValue == path:
            return
        if oldValue is not None:
            self.watcher.removePath(oldValue)
        self._exportDirectory = path
        self.watcher.addPath(self.exportDirectory)
        self.exportDirLabel.setText(self._exportDirectory)
        self.updateNumbers()

    def chooseExportDir(self, givenDir=None):
        state = settings.exportFileDialogState()
        dialog = QFileDialog(self)
        if state:
            dialog.restoreState(state)
        dialogDir = dialog.directory()
        if givenDir is not None:
            dialog.setDirectory(givenDir)
        elif dialogDir is None:
            dialog.setDirectory(
                QStandardPaths.standardLocations(QStandardPaths.DocumentsLocation)[0]
            )
        dialog.setAcceptMode(QFileDialog.AcceptOpen)
        dialog.setFileMode(QFileDialog.Directory)
        ok = dialog.exec_()
        exportDir = QDir.toNativeSeparators(dialog.directory().absolutePath())
        if givenDir is not None:
            dialog.setDirectory(dialogDir)
        settings.setExportFileDialogState(dialog.saveState())
        if ok:
            self.exportDirectory = exportDir

    def updateExportStatus(self):
        value = self.exportBox.isChecked()
        for w in (
            self.exportIcon,
            self.exportDirLabel,
            self.exportDirButton,
            self.warningIcon,
            self.warningLabel,
        ):
            w.setEnabled(value)

    def updateNumbers(self):
        formatOptions = self.formatBtnSet.selectedOptions()
        compressionOptions = self.compressionBtnSet.selectedOptions()
        # number label
        count = len(formatOptions) * len(compressionOptions)
        self.numberLabel.setText(
            self.tr(f"×%n font(s) with base name: {self.baseName}*", n=count)
        )
        # overwrite status
        # XXX: not DRY with the TFont.export logic
        count = 0
        # make a list out of this, otherwise we'll consume the iterator
        compressions = list(map(str.lower, compressionOptions))
        for format in map(str.lower, formatOptions):
            filePath = os.path.join(self.exportDirectory, f"{self.baseName}.{format}")
            for compression in compressions:
                fullPath = filePath
                if compression != "none":
                    fullPath += f".{compression}"
                count += os.path.exists(fullPath)
        visible = bool(count)
        self.warningIcon.setVisible(visible)
        self.warningLabel.setVisible(visible)
        if visible:
            self.warningLabel.setText(
                self.tr("%n file(s) will be overwritten.", n=count)
            )

    def finish(self):
        self.accept()
        export = self.exportBox.isChecked()
        # here we pick a directory, but it won't become the default export dir
        if not export:
            self.chooseExportDir()

    @classmethod
    def getExportParameters(cls, parent, font):
        dialog = cls(font, parent)
        result = dialog.exec_()
        params = dict(
            baseName=dialog.baseName,
            formats=dialog.formatBtnSet.selectedOptions(),
            compression=dialog.compressionBtnSet.selectedOptions(),
            exportDirectory=dialog.exportDirectory,
            removeOverlap=dialog.removeOverlapBox.isChecked(),
            autohint=dialog.autohintBox.isChecked(),
        )
        return (params, result)

    # ----------
    # Qt methods
    # ----------

    def accept(self):
        self.writeSettings()
        super().accept()

    def reject(self):
        self.writeSettings()
        super().reject()

    def closeEvent(self, event):
        super().closeEvent(event)
        if event.isAccepted():
            self.writeSettings()



from defcon import Anchor, Component, Glyph, Guideline
from fontTools.pens.basePen import decomposeQuadraticSegment
from PyQt5.QtCore import QPointF, QRectF, Qt
from PyQt5.QtGui import QColor, QPainter, QPainterPath, QPalette
from PyQt5.QtWidgets import (
    QApplication,
    QMenu,
    QRubberBand,
    QStyle,
    QStyleOptionRubberBand,
)

from trufont.controls.glyphDialogs import AddComponentDialog, EditDialog
from trufont.drawingTools.baseTool import BaseTool
from trufont.tools import bezierMath, platformSpecific
from trufont.tools.uiMethods import (
    maybeProjectUISmoothPointOffcurve,
    moveUIGlyphElements,
    unselectUIGlyphElements,
)

arrowKeys = (Qt.Key_Left, Qt.Key_Up, Qt.Key_Right, Qt.Key_Down)
navKeys = (Qt.Key_Less, Qt.Key_Greater)

_path = QPainterPath()
_path.moveTo(15.33, 4.18)
_path.lineTo(12.36, 11.2)
_path.lineTo(8.37, 7.46)
_path.lineTo(8.37, 23.87)
_path.lineTo(20.85, 12.05)
_path.lineTo(15.33, 12.05)
_path.lineTo(18.17, 5.36)
_path.closeSubpath()


def _pointWithinThreshold(x, y, curve, eps):
    """
    See whether *(x, y)* is within *eps* of *curve*.
    """
    path = QPainterPath()
    path.addEllipse(x - eps, y - eps, 2 * eps, 2 * eps)
    curvePath = QPainterPath()
    if curve[-1].segmentType == "curve":
        p1, p2, p3, p4 = curve
        curvePath.moveTo(p1.x, p1.y)
        curvePath.cubicTo(p2.x, p2.y, p3.x, p3.y, p4.x, p4.y)
        curvePath.cubicTo(p3.x, p3.y, p2.x, p2.y, p1.x, p1.y)
    else:
        first = curve[0]
        curvePath.moveTo(first.x, first.y)
        # PACK for fontTools
        pts = []
        for pt in curve:
            pts.append((pt.x, pt.y))
        # draw
        for pt1, pt2 in decomposeQuadraticSegment(pts[1:]):
            curvePath.quadTo(*pt1 + pt2)
        for pt1, pt2 in decomposeQuadraticSegment(list(reversed(pts[:-1]))):
            curvePath.quadTo(*pt1 + pt2)
    return path.intersects(curvePath)


class SelectionTool(BaseTool):
    icon = _path
    name = QApplication.translate("SelectionTool", "Select")
    shortcut = "V"

    def __init__(self, parent=None):
        super().__init__(parent)
        self._origin = None
        self._mouseItem = None
        self._oldPath = None
        self._oldSelection = set()
        self._rubberBandRect = None
        self._shouldMove = False

    # helpers

    def _createAnchor(self, *_):
        widget = self.parent()
        glyph = self._glyph
        pos = widget.mapToCanvas(widget.mapFromGlobal(self._cachedPos))
        # remove template anchors
        for anchor in glyph.anchors:
            if anchor.name == "new anchor":
                glyph.removeAnchor(anchor)
        # add one at position
        anchor = dict(x=pos.x(), y=pos.y(), name="new anchor")
        glyph.appendAnchor(anchor)
        self._editItem(glyph.anchors[-1])

    def _createComponent(self, *_):
        widget = self.parent()
        glyph = self._glyph
        newGlyph, ok = AddComponentDialog.getNewGlyph(widget, glyph)
        if ok and newGlyph is not None:
            component = glyph.instantiateComponent()
            component.baseGlyph = newGlyph.name
            glyph.appendComponent(component)

    def _createGuideline(self, *_):
        widget = self.parent()
        glyph = self._glyph
        pos = widget.mapToCanvas(widget.mapFromGlobal(self._cachedPos))
        guideline = dict(x=pos.x(), y=pos.y(), angle=0)
        glyph.appendGuideline(guideline)

    def _goToGlyph(self, glyphName):
        widget = self.parent()
        font = self._glyph.font
        if glyphName in font:
            glyph = font[glyphName]
            fontWindow = widget.window()
            fontWindow.openGlyphTab(glyph)

    def _toggleGuideline(self, guideline):
        glyph = self._glyph
        font = glyph.font
        if font is None:
            return
        if isinstance(guideline.getParent(), Glyph):
            glyph.removeGuideline(guideline)
            font.appendGuideline(guideline)
        else:
            font.removeGuideline(guideline)
            glyph.appendGuideline(guideline)

    def _getSelectedCandidatePoint(self):
        """
        If there is exactly one point selected in the glyph, return it.

        Else return None.
        """

        candidates = set()
        for contour in self._glyph:
            sel = contour.selection
            if len(sel) > 1:
                return None
            elif not len(sel):
                continue
            pt = next(iter(sel))
            candidates.add((pt, contour))
        if len(candidates) == 1:
            return next(iter(candidates))
        return None

    def _findSegmentUnderMouse(self, pos, action=None):
        scale = self.parent().inverseScale()
        for contour in self._glyph:
            for index, point in enumerate(contour):
                if point.segmentType == "line":
                    prev = contour.getPoint(index - 1)
                    dist = bezierMath.lineDistance(
                        prev.x, prev.y, point.x, point.y, pos.x(), pos.y()
                    )
                    # TODO: somewhat arbitrary
                    if dist <= 3 * scale:
                        return [prev, point], contour
                elif point.segmentType in ("curve", "qcurve"):
                    if action == "insert":
                        continue
                    if point.segmentType == "curve":
                        bez = [contour.getPoint(index - 3 + i) for i in range(4)]
                    else:
                        bez = [point]
                        i = 1
                        while i < 2 or point.segmentType is None:
                            point = contour.getPoint(index - i)
                            bez.append(point)
                            i += 1
                        bez.reverse()
                    if _pointWithinThreshold(pos.x(), pos.y(), bez, 5 * scale):
                        return bez, contour
        return None

    def _performSegmentClick(self, pos, action=None, segmentTuple=None):
        if segmentTuple is None:
            segmentTuple = self._findSegmentUnderMouse(pos, action)
        if segmentTuple is None:
            return False
        segment, contour = segmentTuple
        prev, point = segment[0], segment[-1]
        if point.segmentType == "line":
            if action == "insert":
                index = contour.index(point) or len(contour)
                contour.holdNotifications()
                for i, t in enumerate((0.35, 0.65)):
                    xt = prev.x + t * (point.x - prev.x)
                    yt = prev.y + t * (point.y - prev.y)
                    contour.insertPoint(index + i, point.__class__((xt, yt)))
                point.segmentType = "curve"
                contour.releaseHeldNotifications()
                return True
        elif point.segmentType not in ("curve", "qcurve"):
            return True
        if action == "selectContour":
            contour.selected = not contour.selected
            self._shouldMove = True
        return True

    def _maybeJoinContour(self, pos):
        if self._mouseItem is None or not isinstance(self._mouseItem, tuple):
            return
        contour, index = self._mouseItem
        point = contour[index]
        if not (point.segmentType and contour.open):
            return
        if index not in (0, len(contour) - 1):
            return
        widget = self.parent()
        items = widget.itemsAt(pos)
        for c, i in items["points"]:
            p = c[i]
            if p == point or not (p.segmentType and c.open):
                continue
            if i not in (0, len(c) - 1):
                continue
            if c != contour:
                # TODO: blacklist single onCurve contours
                # Note reverse uses different point objects
                # TODO: does it have to work this way?
                # update: I think frederik changed that in master defcon
                if not index:
                    contour.reverse()
                if i == -1:
                    c.reverse()
                dragPoint = contour[-1]
                contour.removePoint(dragPoint)
                c[0].segmentType = dragPoint.segmentType
                c.drawPoints(contour)
                glyph = c.glyph
                glyph.removeContour(c)
                contour.dirty = True
            else:
                if point.segmentType == "move":
                    point.x = p.x
                    point.y = p.y
                    c.removePoint(p)
                else:
                    c.removePoint(point)
                c[0].segmentType = "line"
                c.dirty = True
            return

    def _moveForEvent(self, event):
        key = event.key()
        modifiers = event.modifiers()
        dx, dy = 0, 0
        if key == Qt.Key_Left:
            dx = -1
        elif key == Qt.Key_Up:
            dy = 1
        elif key == Qt.Key_Right:
            dx = 1
        elif key == Qt.Key_Down:
            dy = -1
        if modifiers & Qt.ShiftModifier:
            dx *= 10
            dy *= 10
            if modifiers & Qt.ControlModifier:
                dx *= 10
                dy *= 10
        return (dx, dy)

    def _editItem(self, item):
        widget = self.parent()
        newName, x, y, ok = EditDialog.getNewProperties(widget, item)
        if ok:
            item.name = newName
            item.x = x
            item.y = y

    def _reverse(self, target=None):
        if target is None:
            selectedContours = set()
            for contour in self._glyph:
                if contour.selection:
                    selectedContours.add(contour)
            target = selectedContours or self._glyph
        if not target:
            return
        self._glyph.holdNotifications()
        for contour in target:
            contour.reverse()
        self._glyph.releaseHeldNotifications()

    def _setStartPoint(self, contour, index):
        if not index:
            return
        contour.setStartPoint(index)

    # events

    def contextMenuEvent(self, event):
        widget = self.parent()
        self._cachedPos = event.globalPos()
        menu = QMenu(widget)
        item = widget.itemAt(event.localPos())
        targetContour = None
        if item is not None:
            if isinstance(item, tuple):
                contour, index = item
                targetContour = contour
                if True or contour[index].segmentType is not None:  # XXX
                    menu.addAction(
                        self.tr("Set Start Point"),
                        lambda: self._setStartPoint(contour, index),
                    )
            elif isinstance(item, Component):
                menu.addAction(
                    self.tr("Go To Glyph"), lambda: self._goToGlyph(item.baseGlyph)
                )
                menu.addAction(
                    self.tr("Decompose"), lambda: self._glyph.decomposeComponent(item)
                )
                menu.addAction(
                    self.tr("Decompose All"), self._glyph.decomposeAllComponents
                )
            elif isinstance(item, Guideline):
                if isinstance(item.getParent(), Glyph):
                    text = self.tr("Make Global")
                else:
                    text = self.tr("Make Local")
                menu.addAction(text, lambda: self._toggleGuideline(item))
        if len(self._glyph):
            if targetContour is not None:
                reverseText = self.tr("Reverse Contour")
                target = [targetContour]
            else:
                # XXX: text and action shouldnt be decoupled
                if self._glyph.selection:
                    reverseText = self.tr("Reverse Selected Contours")
                else:
                    reverseText = self.tr("Reverse All Contours")
                target = None
            menu.addAction(reverseText, lambda: self._reverse(target))
            menu.addSeparator()
        if len(self._glyph.layer) > 1:
            menu.addAction(self.tr("Add Component…"), self._createComponent)
        menu.addAction(self.tr("Add Anchor"), self._createAnchor)
        menu.addAction(self.tr("Add Guideline"), self._createGuideline)
        menu.exec_(self._cachedPos)
        self._cachedPos = None

    def keyPressEvent(self, event):
        key = event.key()
        # XXX: this shouldn't be tool-specific!
        if key in arrowKeys:
            dx, dy = self._moveForEvent(event)
            modifiers = event.modifiers()
            kwargs = dict()
            if modifiers == platformSpecific.combinedModifiers():
                kwargs["nudgePoints"] = True
            elif modifiers & Qt.AltModifier:
                kwargs["slidePoints"] = True
            moveUIGlyphElements(self._glyph, dx, dy, **kwargs)
        # TODO: nav shouldn't be specific to this tool
        elif key in navKeys:
            pack = self._getSelectedCandidatePoint()
            if pack is not None:
                point, contour = pack
                point.selected = False
                index = contour.index(point)
                offset = int(key == Qt.Key_Greater) or -1
                newPoint = contour.getPoint(index + offset)
                newPoint.selected = True
                contour.postNotification(notification="Contour.SelectionChanged")
        elif key == Qt.Key_Return:
            self._glyph.beginUndoGroup()
            for contour in self._glyph:
                for index, point in enumerate(contour):
                    changed = False
                    if point.segmentType is not None and point.selected:
                        if all(
                            contour.getPoint(index + d).segmentType for d in (-1, 1)
                        ):
                            continue
                        point.smooth = not point.smooth
                        changed = True
                    if changed:
                        contour.dirty = True
            self._glyph.endUndoGroup()

    def mousePressEvent(self, event):
        if event.button() != Qt.LeftButton:
            super().mousePressEvent(event)
            return
        widget = self.parent()
        addToSelection = event.modifiers() & Qt.ControlModifier
        self._glyph.beginUndoGroup()
        self._origin = self._prevPos = pos = self.magnetPos(event.localPos())
        self._mouseItem = widget.itemAt(self._origin)
        if self._mouseItem is not None:
            item = self._mouseItem
            contour = None
            if isinstance(item, tuple):
                contour, i = item
                item = contour[i]
            if not (item.selected or addToSelection):
                unselectUIGlyphElements(self._glyph)
            if addToSelection:
                item.selected = not item.selected
            else:
                item.selected = True
            if contour is not None:
                contour.postNotification(notification="Contour.SelectionChanged")
        else:
            action = "insert" if event.modifiers() & Qt.AltModifier else None
            segmentTuple = self._findSegmentUnderMouse(pos, action)
            if segmentTuple is not None:
                segment, contour = segmentTuple
                selected = segment[0].selected and segment[-1].selected
            else:
                selected = False
            if not selected:
                if addToSelection:
                    self._oldSelection = self._glyph.selection
                else:
                    unselectUIGlyphElements(self._glyph)
                self._performSegmentClick(pos, action, segmentTuple)
            else:
                self._shouldMove = True
        if self._mouseItem is not None or self._shouldMove:
            self._oldPath = self._glyph.getRepresentation(
                "defconQt.NoComponentsQPainterPath"
            )
        widget.update()

    def mouseMoveEvent(self, event):
        if not event.buttons() & Qt.LeftButton:
            super().mouseMoveEvent(event)
            return
        if self._origin is None:
            return
        glyph = self._glyph
        widget = self.parent()
        if self._shouldMove or self._mouseItem is not None:
            canvasPos = event.pos()
            modifiers = event.modifiers()
            if modifiers & Qt.ShiftModifier:
                # we clamp to the mouseDownPos, unless we have a
                # single offCurve in which case we clamp it against
                # its parent
                canvasPos = self.clampToOrigin(canvasPos, self._origin)
                if isinstance(self._mouseItem, tuple):
                    selection = glyph.selection
                    if len(selection) == 1:
                        c, i = self._mouseItem
                        p_ = c[i]
                        if p_.segmentType is None:
                            for d in (-1, 1):
                                p__ = c.getPoint(i + d)
                                if p__.segmentType is not None:
                                    canvasPos = self.clampToOrigin(
                                        canvasPos, QPointF(p__.x, p__.y)
                                    )
                                    p_.x = canvasPos.x()
                                    p_.y = canvasPos.y()
                                    c.dirty = True
                                    return
            dx = canvasPos.x() - self._prevPos.x()
            dy = canvasPos.y() - self._prevPos.y()
            kwargs = dict()
            if modifiers == platformSpecific.combinedModifiers():
                kwargs["nudgePoints"] = True
            elif modifiers & Qt.AltModifier:
                kwargs["slidePoints"] = True
            moveUIGlyphElements(glyph, dx, dy, **kwargs)
            self._prevPos = canvasPos
        else:
            canvasPos = event.localPos()
            self._rubberBandRect = QRectF(self._origin, canvasPos).normalized()
            items = widget.items(self._rubberBandRect)
            points = {c[i] for c, i in items["points"]}
            if event.modifiers() & Qt.ControlModifier:
                points ^= self._oldSelection
            # TODO: fine-tune this more, maybe add optional args to items...
            if event.modifiers() & Qt.AltModifier:
                points = {pt for pt in points if pt.segmentType}
            if points != self._glyph.selection:
                # TODO: doing this takes more time than by-contour
                # discrimination for large point count
                glyph.selection = points
        widget.update()

    def mouseReleaseEvent(self, event):
        if event.button() == Qt.LeftButton:
            # we changed glyph during this mouse press, skip special
            # processing,
            # (this would be a no-op except for unbalanced undo warning)
            if hasattr(self, "_switched"):
                del self._switched
            else:
                self._maybeJoinContour(event.localPos())
                self._glyph.endUndoGroup()
            self._mouseItem = None
            self._oldPath = None
            self._oldSelection = set()
            self._rubberBandRect = None
            self._shouldMove = False
            self.parent().update()
        else:
            super().mouseReleaseEvent(event)

    def mouseDoubleClickEvent(self, event):
        widget = self.parent()
        self._mouseItem = item = widget.itemAt(self._origin)
        self._glyph.beginUndoGroup()
        if item is not None:
            if isinstance(item, tuple):
                contour, index = item
                point = contour[index]
                if point.segmentType is not None:
                    if all(contour.getPoint(index + d).segmentType for d in (-1, 1)):
                        return
                    point.smooth = not point.smooth
                    contour.dirty = True
                    # if we have one offCurve, make it tangent
                    maybeProjectUISmoothPointOffcurve(contour, index)
            elif isinstance(item, (Anchor, Guideline)):
                self._editItem(item)
        else:
            if self._performSegmentClick(event.localPos(), "selectContour"):
                return
            index = widget.indexForPoint(widget.mapFromCanvas(event.localPos()))
            if index is not None:
                # we're about to switch glyph, end undo group first
                self._glyph.endUndoGroup()
                self._switched = True
                widget.setActiveIndex(index)
        # don't perform move events on double click
        self._origin = None

    # custom painting

    def paintBackground(self, painter, index):
        if index != self.parent().activeIndex():
            return
        if self._oldPath is not None:
            # XXX: honor partialAliasing
            painter.save()
            pen = painter.pen()
            pen.setColor(QColor(210, 210, 210))
            pen.setWidth(0)
            painter.setPen(pen)
            painter.drawPath(self._oldPath)
            painter.restore()

    def paint(self, painter, index):
        if self._rubberBandRect is None:
            return
        widget = self.parent()
        if index != widget.activeIndex():
            return
        rect = self._rubberBandRect
        if platformSpecific.useBuiltinRubberBand():
            # okay, OS-native rubber band does not support painting with
            # floating-point coordinates
            # paint directly on the widget with unscaled context
            widgetOrigin = widget.mapFromCanvas(rect.bottomLeft())
            widgetMove = widget.mapFromCanvas(rect.topRight())
            option = QStyleOptionRubberBand()
            option.initFrom(widget)
            option.opaque = False
            option.rect = QRectF(widgetOrigin, widgetMove).toRect()
            option.shape = QRubberBand.Rectangle
            painter.save()
            painter.setRenderHint(QPainter.Antialiasing, False)
            painter.resetTransform()
            widget.style().drawControl(QStyle.CE_RubberBand, option, painter, widget)
            painter.restore()
        else:
            highlight = widget.palette().color(QPalette.Active, QPalette.Highlight)
            painter.save()
            painter.setRenderHint(QPainter.Antialiasing, False)
            pen = painter.pen()
            pen.setColor(highlight.darker(120))
            pen.setWidth(0)
            painter.setPen(pen)
            highlight.setAlphaF(0.35)
            painter.setBrush(highlight)
            painter.drawRect(rect)
            painter.restore()
